# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
# SPDX-FileCopyrightText: © 2008-2022 Taylor Raak
# SPDX-FileCopyrightText: © 2024 Christian BUHTZ <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
from PyQt6.QtWidgets import (QDialog,
                             QVBoxLayout,
                             QHBoxLayout,
                             QGridLayout,
                             QLabel,
                             QSpinBox,
                             QLineEdit,
                             QCheckBox)
import config
import tools
import qttools
import messagebox
from manageprofiles.statebindcheckbox import StateBindCheckBox


class ExpertOptionsTab(QDialog):
    """The 'Expert Options' tab in the Manage Profiles dialog."""

    def __init__(self, parent):
        super().__init__(parent=parent)

        self._parent_dialog = parent

        tab_layout = QVBoxLayout(self)

        label = QLabel('<strong>{}</strong> {}'.format(
            _('Caution:'),
            _('These options are for advanced configurations. Modify '
              'only if fully aware of their implications.')))
        label.setWordWrap(True)
        tab_layout.addWidget(label)

        # --- rsync with nice ---
        tab_layout.addWidget(QLabel(
            _("Run 'rsync' with '{cmd}':").format(cmd='nice')))

        grid = QGridLayout()
        grid.setColumnMinimumWidth(0, 20)  # left indent
        tab_layout.addLayout(grid)

        self.cbNiceOnCron = QCheckBox(
            _('as cron job')
            + self._default_string(self.config.DEFAULT_RUN_NICE_FROM_CRON),
            self)
        grid.addWidget(self.cbNiceOnCron, 0, 1)

        self.cbNiceOnRemote = QCheckBox(
            _('on remote host')
            + self._default_string(self.config.DEFAULT_RUN_NICE_ON_REMOTE),
            self)
        grid.addWidget(self.cbNiceOnRemote, 1, 1)

        # --- rsync with ionice ---
        tab_layout.addWidget(QLabel(
            _("Run 'rsync' with '{cmd}':").format(cmd='ionice')))
        grid = QGridLayout()
        grid.setColumnMinimumWidth(0, 20)
        tab_layout.addLayout(grid)

        self.cbIoniceOnCron = QCheckBox(
            _('as cron job')
            + self._default_string(self.config.DEFAULT_RUN_IONICE_FROM_CRON),
            self)
        grid.addWidget(self.cbIoniceOnCron, 0, 1)

        self.cbIoniceOnUser = QCheckBox(
            _('when taking a manual snapshot')
            + self._default_string(self.config.DEFAULT_RUN_IONICE_FROM_USER),
            self)
        grid.addWidget(self.cbIoniceOnUser, 1, 1)

        self.cbIoniceOnRemote = QCheckBox(
            _('on remote host')
            + self._default_string(self.config.DEFAULT_RUN_IONICE_ON_REMOTE),
            self)
        grid.addWidget(self.cbIoniceOnRemote, 2, 1)

        # --- rsync with nocache ---
        tab_layout.addWidget(QLabel(
            _("Run 'rsync' with '{cmd}':").format(cmd='nocache')))

        grid = QGridLayout()
        grid.setColumnMinimumWidth(0, 20)
        tab_layout.addLayout(grid)

        nocache_available = tools.checkCommand('nocache')
        if not nocache_available:
            grid.addWidget(
                QLabel(
                    '<em>'
                    + _("Please install 'nocache' to enable this option.")
                    + '</em>'),
                0,
                1)

        self.cbNocacheOnLocal = QCheckBox(
            _('on local machine')
            + self._default_string(self.config.DEFAULT_RUN_NOCACHE_ON_LOCAL),
            self)
        grid.addWidget(self.cbNocacheOnLocal, 1, 1)
        self.cbNocacheOnLocal.setEnabled(nocache_available)

        self.cbNocacheOnRemote = QCheckBox(
            _('on remote host')
            + self._default_string(self.config.DEFAULT_RUN_NOCACHE_ON_REMOTE),
            self)
        grid.addWidget(self.cbNocacheOnRemote, 2, 1)

        # --- redirect output ---
        self.cbRedirectStdoutInCron = QCheckBox(
            _('Redirect stdout to /dev/null in cronjobs.')
            + self._default_string(
                self.config.DEFAULT_REDIRECT_STDOUT_IN_CRON),
            self)
        qttools.set_wrapped_tooltip(
            self.cbRedirectStdoutInCron,
            _('Cron will automatically send an email with attached output '
              'of cronjobs if an MTA is installed.')
        )
        tab_layout.addWidget(self.cbRedirectStdoutInCron)

        self.cbRedirectStderrInCron = QCheckBox(
            _('Redirect stderr to /dev/null in cronjobs.')
            + self._default_string(
                self.config.DEFAULT_REDIRECT_STDERR_IN_CRON),
            self)
        qttools.set_wrapped_tooltip(
            self.cbRedirectStderrInCron,
            _('Cron will automatically send an email with attached errors '
              'of cronjobs if an MTA is installed.')
        )
        tab_layout.addWidget(self.cbRedirectStderrInCron)

        # bandwidth limit
        hlayout = QHBoxLayout()
        tab_layout.addLayout(hlayout)

        self.spbBwlimit = QSpinBox(self)
        self.spbBwlimit.setSuffix(' ' + _('KB/sec'))
        self.spbBwlimit.setSingleStep(100)
        self.spbBwlimit.setRange(0, 1000000)

        self.cbBwlimit = StateBindCheckBox(
            _('Limit rsync bandwidth usage:'), self, self.spbBwlimit)
        hlayout.addWidget(self.cbBwlimit)
        hlayout.addWidget(self.spbBwlimit)
        hlayout.addStretch()

        qttools.set_wrapped_tooltip(
            self.cbBwlimit,
            [
                "Uses 'rsync --bwlimit=RATE'. From 'man rsync':",
                'This option allows you to specify the maximum transfer rate '
                'for the data sent over the socket, specified in units per '
                'second. The RATE value can be suffixed with a string to '
                'indicate a size multiplier, and may be a fractional value '
                '(e.g. "--bwlimit=1.5m").',
                'If no suffix is specified, the value will be assumed to be '
                'in units of 1024 bytes (as if "K" or "KiB" had been '
                'appended).',
                'See the --max-size option for a description of all the '
                'available suffixes. A value of zero specifies no limit.'
                '',
                'For backward-compatibility reasons, the rate limit will be '
                'rounded to the nearest KiB unit, so no rate smaller than '
                '1024 bytes per second is possible.',
                '',
                'Rsync writes data over the socket in blocks, and this option '
                'both limits the size of the blocks that rsync writes, and '
                'tries to keep the average transfer rate at the requested '
                'limit. Some "burstiness" may be seen where rsync writes out '
                'a block of data and then sleeps to bring the average rate '
                'into compliance.',
                '',
                'Due to the internal buffering of data, the --progress '
                'option may not be an accurate reflection on how fast the '
                'data is being sent. This is because some files can show up '
                'as being rapidly sent when the data is quickly buffered, '
                'while other can show up as very slow when the flushing of '
                'the output buffer occurs. This may be fixed in a future '
                'version.'
            ]
        )

        self.cbPreserveAcl = QCheckBox(_('Preserve ACL'), self)
        qttools.set_wrapped_tooltip(
            self.cbPreserveAcl,
            [
                "Uses 'rsync -A'. From 'man rsync':",
                'This option causes rsync to update the destination ACLs to '
                'be the same as the source ACLs. The option also implies '
                '--perms.',
                '',
                'The source and destination systems must have compatible ACL '
                'entries for this option to work properly. See the '
                '--fake-super option for a way to backup and restore ACLs '
                'that are not compatible.'
            ]
        )
        tab_layout.addWidget(self.cbPreserveAcl)

        self.cbPreserveXattr = QCheckBox(
            _('Preserve extended attributes (xattr)'), self)
        qttools.set_wrapped_tooltip(
            self.cbPreserveXattr,
            [
                "Uses 'rsync -X'. From 'man rsync':",
                'This option causes rsync to update the destination extended '
                'attributes to be the same as the source ones.',
                '',
                'For systems that support extended-attribute namespaces, a '
                'copy being done by a super-user copies all namespaces '
                'except system.*. A normal user only copies the user.* '
                'namespace. To be able to backup and restore non-user '
                'namespaces as a normal user, see the --fake-super option.',
                '',
                'Note that this option does not copy rsyncs special xattr '
                'values (e.g. those used by --fake-super) unless you repeat '
                'the option (e.g. -XX). This "copy all xattrs" mode cannot be '
                'used with --fake-super.'
            ]
        )
        tab_layout.addWidget(self.cbPreserveXattr)

        self.cbCopyUnsafeLinks = QCheckBox(
            _('Copy unsafe links (works only with absolute links)'), self)
        qttools.set_wrapped_tooltip(
            self.cbCopyUnsafeLinks,
            [
                "Uses 'rsync --copy-unsafe-links'. From 'man rsync':",
                'This tells rsync to copy the referent of symbolic links that '
                'point outside the copied tree. Absolute symlinks are also '
                'treated like ordinary files, and so are any symlinks in the '
                'source path itself when --relative is used. This option has '
                'no additional effect if --copy-links was also specified.'
            ]
        )
        tab_layout.addWidget(self.cbCopyUnsafeLinks)

        self.cbCopyLinks = QCheckBox(
            _('Copy links (dereference symbolic links)'), self)
        qttools.set_wrapped_tooltip(
            self.cbCopyLinks,
            [
                "Uses 'rsync --copy-links'. From 'man rsync':",
                'When symlinks are encountered, the item that they point to '
                '(the referent) is copied, rather than the symlink. In older '
                'versions of rsync, this option also had the side-effect of '
                'telling the receiving side to follow symlinks, such as '
                'symlinks to directories. In a modern rsync such as this one,'
                ' you will need to specify --keep-dirlinks (-K) to get this '
                'extra behavior. The only exception is when sending files to '
                'an rsync that is too old to understand -K -- in that case, '
                'the -L option will still have the side-effect of -K on that '
                'older receiving rsync.'
            ]
        )
        tab_layout.addWidget(self.cbCopyLinks)

        # one file system option
        self.cbOneFileSystem = QCheckBox(
            _('Restrict to one file system'), self)
        qttools.set_wrapped_tooltip(
            self.cbOneFileSystem,
            [
                "Uses 'rsync --one-file-system'. From 'man rsync':",
                'This tells rsync to avoid crossing a filesystem boundary '
                'when recursing. This does not limit the user\'s ability '
                'to specify items to copy from multiple filesystems, just '
                'rsync\'s recursion through the hierarchy of each directory '
                'that the user specified, and also the analogous recursion '
                'on the receiving side during deletion. Also keep in mind '
                'that rsync treats a "bind" mount to the same device as '
                'being on the same filesystem.'
            ]
        )
        tab_layout.addWidget(self.cbOneFileSystem)

        # additional rsync options
        tooltip = _('Options must be quoted e.g. {example}.').format(
            example='--exclude-from="/path/to/my exclude file"')

        self.txtRsyncOptions = QLineEdit(self)
        self.txtRsyncOptions.editingFinished.connect(
            self._slot_rsync_options_editing_finished)
        self.txtRsyncOptions.setToolTip(tooltip)

        self.cbRsyncOptions = StateBindCheckBox(
            _('Paste additional options to rsync'),
            self,
            self.txtRsyncOptions)

        self.cbRsyncOptions.setToolTip(tooltip)

        # ssh prefix
        tooltip = [
            _('Prefix to run before every command on remote host.'),
            _("Variables need to be escaped with \\$FOO. This doesn't touch "
              'rsync. So to add a prefix for rsync use "{example_value}" with '
              '{rsync_options_value}.').format(
                  example_value=self.cbRsyncOptions.text(),
                  rsync_options_value \
                      ='--rsync-path="FOO=bar:\\$FOO /usr/bin/rsync"'),
            '',
            '{default}: {def_value}'.format(
                default=_('default'),
                def_value=self.config.DEFAULT_SSH_PREFIX)
        ]
        self.txtSshPrefix = QLineEdit(self)
        qttools.set_wrapped_tooltip(self.txtSshPrefix, tooltip)
        self.cbSshPrefix = StateBindCheckBox(
            _('Add prefix to SSH commands'), self, self.txtSshPrefix)
        qttools.set_wrapped_tooltip(self.cbSshPrefix, tooltip)

        sub_grid = QGridLayout()
        sub_grid.addWidget(self.cbRsyncOptions, 0, 0)
        sub_grid.addWidget(self.txtRsyncOptions, 0, 1)
        sub_grid.addWidget(self.cbSshPrefix, 1, 0)
        sub_grid.addWidget(self.txtSshPrefix, 1, 1)
        tab_layout.addLayout(sub_grid)

        self.cbSshCheckPing = QCheckBox(_('Check if remote host is online'))
        qttools.set_wrapped_tooltip(
            self.cbSshCheckPing,
            _('Warning: If disabled and the remote host is not available, '
              'this could lead to some weird errors.')
        )
        self.cbSshCheckCommands = QCheckBox(
            _('Check if remote host supports all necessary commands.'))
        qttools.set_wrapped_tooltip(
            self.cbSshCheckCommands,
            _('Warning: If disabled and the remote host does not support all '
              'necessary commands, this could lead to some weird errors.')
        )
        tab_layout.addWidget(self.cbSshCheckPing)
        tab_layout.addWidget(self.cbSshCheckCommands)

        #
        tab_layout.addStretch()


    @property
    def config(self) -> config.Config:
        return self._parent_dialog.config

    def _default_string(self, value: bool) -> str:
        return ' ' + _('(default: {})').format(
            _('enabled') if value else _('disabled'))

    def load_values(self):
        self.cbNiceOnCron.setChecked(self.config.niceOnCron())
        self.cbIoniceOnCron.setChecked(self.config.ioniceOnCron())
        self.cbIoniceOnUser.setChecked(self.config.ioniceOnUser())
        self.cbNiceOnRemote.setChecked(self.config.niceOnRemote())
        self.cbIoniceOnRemote.setChecked(self.config.ioniceOnRemote())
        self.cbNocacheOnLocal.setChecked(
            self.config.nocacheOnLocal() and self.cbNocacheOnLocal.isEnabled())
        self.cbNocacheOnRemote.setChecked(self.config.nocacheOnRemote())
        self.cbRedirectStdoutInCron.setChecked(
            self.config.redirectStdoutInCron())
        self.cbRedirectStderrInCron.setChecked(
            self.config.redirectStderrInCron())
        self.cbBwlimit.setChecked(self.config.bwlimitEnabled())
        self.spbBwlimit.setValue(self.config.bwlimit())
        self.cbPreserveAcl.setChecked(self.config.preserveAcl())
        self.cbPreserveXattr.setChecked(self.config.preserveXattr())
        self.cbCopyUnsafeLinks.setChecked(self.config.copyUnsafeLinks())
        self.cbCopyLinks.setChecked(self.config.copyLinks())
        self.cbOneFileSystem.setChecked(self.config.oneFileSystem())
        self.cbRsyncOptions.setChecked(self.config.rsyncOptionsEnabled())
        self.txtRsyncOptions.setText(self.config.rsyncOptions())
        self.cbSshPrefix.setChecked(self.config.sshPrefixEnabled())
        self.txtSshPrefix.setText(self.config.sshPrefix())
        self.cbSshCheckPing.setChecked(self.config.sshCheckPingHost())
        self.cbSshCheckCommands.setChecked(self.config.sshCheckCommands())

    def store_values(self):
        self.config.setNiceOnCron(self.cbNiceOnCron.isChecked())
        self.config.setIoniceOnCron(self.cbIoniceOnCron.isChecked())
        self.config.setIoniceOnUser(self.cbIoniceOnUser.isChecked())
        self.config.setNiceOnRemote(self.cbNiceOnRemote.isChecked())
        self.config.setIoniceOnRemote(self.cbIoniceOnRemote.isChecked())
        self.config.setNocacheOnLocal(self.cbNocacheOnLocal.isChecked())
        self.config.setNocacheOnRemote(self.cbNocacheOnRemote.isChecked())
        self.config.setRedirectStdoutInCron(
            self.cbRedirectStdoutInCron.isChecked())
        self.config.setRedirectStderrInCron(
            self.cbRedirectStderrInCron.isChecked())
        self.config.setBwlimit(self.cbBwlimit.isChecked(),
                               self.spbBwlimit.value())
        self.config.setPreserveAcl(self.cbPreserveAcl.isChecked())
        self.config.setPreserveXattr(self.cbPreserveXattr.isChecked())
        self.config.setCopyUnsafeLinks(self.cbCopyUnsafeLinks.isChecked())
        self.config.setCopyLinks(self.cbCopyLinks.isChecked())
        self.config.setOneFileSystem(self.cbOneFileSystem.isChecked())
        self.config.setRsyncOptions(self.cbRsyncOptions.isChecked(),
                                    self.txtRsyncOptions.text())
        self.config.setSshPrefix(self.cbSshPrefix.isChecked(),
                                 self.txtSshPrefix.text())
        self.config.setSshCheckPingHost(self.cbSshCheckPing.isChecked())
        self.config.setSshCheckCommands(self.cbSshCheckCommands.isChecked())

    def update_items_state(self, enabled: bool):
        self.cbNiceOnRemote.setEnabled(enabled)
        self.cbIoniceOnRemote.setEnabled(enabled)
        self.cbNocacheOnRemote.setEnabled(enabled)
        self.cbSshPrefix.setVisible(enabled)
        self.txtSshPrefix.setVisible(enabled)
        self.cbSshCheckPing.setVisible(enabled)
        self.cbSshCheckCommands.setVisible(enabled)

    def _slot_rsync_options_editing_finished(self):
        """When editing the rsync options is finished warn and remove
        --old-args option if present.
        """
        txt = self.txtRsyncOptions.text()

        if '--old-args' in txt:
            # No translation for this message because it is a rare case.
            messagebox.warning(
                text='Found rsync flag "--old-args". That flag will be removed'
                ' from the options because it conflicts with the flag "-s" '
                '(also known as "--secluded-args" or "--protected-args") which'
                ' is used by Back In Time to force the "new form of argument '
                'protection" in rsync.',
                widget_to_center_on=self
            )

            # Don't leave two-blank spaces between other arguments
            txt = txt.replace('--old-args ', '')
            txt = txt.replace(' --old-args', '')
            txt = txt.replace('--old-args', '')
            self.txtRsyncOptions.setText(txt)



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
import os
import sys
import unittest
from unittest.mock import patch
from datetime import datetime, timedelta
from tempfile import TemporaryDirectory
from test import generic

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
import snapshots
import mount


class Take(generic.SnapshotsTestCase):
    def setUp(self):
        super().setUp()
        self.include = TemporaryDirectory()
        generic.create_test_files(self.include.name)

    def tearDown(self):
        super().tearDown()
        self.include.cleanup()

    def remount(self):
        # dummy method only used in TestTakeSnapshotSSH
        pass

    def getInode(self, sid):
        return os.stat(sid.pathBackup(os.path.join(self.include.name, 'test'))).st_ino

    @patch('time.sleep')  # speed up unittest
    def test_four_snapshots(self, sleep):
        now = datetime.today() - timedelta(minutes = 6)
        sid1 = snapshots.SID(now, self.cfg)

        # Note: 'self.sn' is of type 'Snapshots'
        # First boolean: Snapshot succeeded
        # Second boolean: Error occurred
        self.assertListEqual(
                [True, False],  # Snapshot without error
                self.sn.takeSnapshot(
                    sid=sid1,
                    now=now,
                    include_folders=[
                        (self.include.name, 0),  # '0' means it is a file
                    ]
                )
        )

        self.assertTrue(sid1.exists())
        self.assertTrue(sid1.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'foo', 'bar', 'baz')))
        self.assertTrue(sid1.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'test')))
        self.assertTrue(sid1.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'file with spaces')))
        self.assertExists(self.cfg.anacronSpoolFile())
        for f in ('config',
                  'fileinfo.bz2',
                  'info',
                  'takesnapshot.log.bz2'):
            self.assertExists(sid1.path(f))

        for f in ('failed',
                  'save_to_continue'):
            self.assertNotExists(sid1.path(f))

        # second takeSnapshot which should not create a new snapshot as nothing
        # has changed
        os.remove(self.cfg.anacronSpoolFile())
        now = datetime.today() - timedelta(minutes = 4)
        sid2 = snapshots.SID(now, self.cfg)

        self.assertListEqual([False, False], self.sn.takeSnapshot(sid2, now, [(self.include.name, 0),]))
        self.assertFalse(sid2.exists())
        self.assertExists(self.cfg.anacronSpoolFile())

        # third takeSnapshot
        self.remount()
        with open(os.path.join(self.include.name, 'lalala'), 'wt') as f:
            f.write('asdf')

        now = datetime.today() - timedelta(minutes = 2)
        sid3 = snapshots.SID(now, self.cfg)

        self.assertListEqual([True, False], self.sn.takeSnapshot(sid3, now, [(self.include.name, 0),]))
        self.assertTrue(sid3.exists())
        self.assertTrue(sid3.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'lalala')))
        inode1 = self.getInode(sid1)
        inode3 = self.getInode(sid3)
        self.assertEqual(inode1, inode3)

        # fourth takeSnapshot with force create new snapshot even if nothing
        # has changed
        self.cfg.setTakeSnapshotRegardlessOfChanges(True)
        now = datetime.today()
        sid4 = snapshots.SID(now, self.cfg)

        self.assertListEqual([True, False], self.sn.takeSnapshot(sid4, now, [(self.include.name, 0),]))
        self.assertTrue(sid4.exists())
        self.assertTrue(sid4.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'foo', 'bar', 'baz')))
        self.assertTrue(sid4.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'test')))

    @patch('time.sleep') # speed up unittest
    def test_spaces_in_include(self, sleep):
        now = datetime.today()
        sid1 = snapshots.SID(now, self.cfg)
        include = os.path.join(self.include.name, 'test path with spaces')
        generic.create_test_files(include)

        self.assertListEqual([True, False], self.sn.takeSnapshot(sid1, now, [(include, 0),]))
        self.assertTrue(sid1.exists())
        self.assertTrue(sid1.isExistingPathInsideSnapshotFolder(os.path.join(include, 'foo', 'bar', 'baz')))
        self.assertTrue(sid1.isExistingPathInsideSnapshotFolder(os.path.join(include, 'test')))
        for f in ('config',
                  'fileinfo.bz2',
                  'info',
                  'takesnapshot.log.bz2'):
            self.assertExists(sid1.path(f))

        for f in ('failed',
                  'save_to_continue'):
            self.assertNotExists(sid1.path(f))

    @patch('time.sleep') # speed up unittest
    def test_exclude(self, sleep):
        now = datetime.today()
        sid1 = snapshots.SID(now, self.cfg)
        self.cfg.setExclude(['bar/baz'])

        self.assertListEqual([True, False], self.sn.takeSnapshot(sid1, now, [(self.include.name, 0),]))
        self.assertTrue(sid1.exists())
        self.assertTrue(sid1.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'foo', 'bar')))
        self.assertFalse(sid1.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'foo', 'bar', 'baz')))
        self.assertTrue(sid1.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'test')))
        for f in ('config',
                  'fileinfo.bz2',
                  'info',
                  'takesnapshot.log.bz2'):
            self.assertExists(sid1.path(f))

        for f in ('failed',
                  'save_to_continue'):
            self.assertNotExists(sid1.path(f))

    @patch('time.sleep') # speed up unittest
    def test_spaces_in_exclude(self, sleep):
        now = datetime.today()
        sid1 = snapshots.SID(now, self.cfg)
        exclude = os.path.join(self.include.name, 'test path with spaces')
        generic.create_test_files(exclude)
        self.cfg.setExclude([exclude])

        self.assertListEqual([True, False], self.sn.takeSnapshot(sid1, now, [(self.include.name, 0),]))
        self.assertTrue(sid1.exists())
        self.assertTrue(sid1.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'foo', 'bar', 'baz')))
        self.assertTrue(sid1.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'test')))
        self.assertFalse(sid1.isExistingPathInsideSnapshotFolder(exclude))
        for f in ('config',
                  'fileinfo.bz2',
                  'info',
                  'takesnapshot.log.bz2'):
            self.assertExists(sid1.path(f))

        for f in ('failed',
                  'save_to_continue'):
            self.assertNotExists(sid1.path(f))

    @patch('time.sleep') # speed up unittest
    def test_error(self, sleep):
        with generic.mockPermissions(os.path.join(self.include.name, 'test')):
            now = datetime.today()
            sid1 = snapshots.SID(now, self.cfg)

            self.assertListEqual([True, True], self.sn.takeSnapshot(sid1, now, [(self.include.name, 0),]))
            self.assertTrue(sid1.exists())
            self.assertTrue(sid1.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'foo', 'bar', 'baz')))
            self.assertFalse(sid1.isExistingPathInsideSnapshotFolder(os.path.join(self.include.name, 'test')))
            for f in ('config',
                      'fileinfo.bz2',
                      'info',
                      'takesnapshot.log.bz2',
                      'failed'):
                self.assertExists(sid1.path(f))
            self.assertNotExists(self.cfg.anacronSpoolFile())

    @patch('time.sleep') # speed up unittest
    def test_error_without_continue(self, sleep):
        with generic.mockPermissions(os.path.join(self.include.name, 'test')):
            self.cfg.setContinueOnErrors(False)
            now = datetime.today()
            sid1 = snapshots.SID(now, self.cfg)

            self.assertListEqual([False, True], self.sn.takeSnapshot(sid1, now, [(self.include.name, 0),]))
            self.assertFalse(sid1.exists())

    @patch('time.sleep') # speed up unittest
    def test_new_exists(self, sleep):
        new_snapshot = snapshots.NewSnapshot(self.cfg)
        new_snapshot.makeDirs()
        with open(new_snapshot.path('leftover'), 'wt') as f:
            f.write('foo')

        now = datetime.today() - timedelta(minutes = 6)
        sid1 = snapshots.SID(now, self.cfg)

        self.assertListEqual([True, False], self.sn.takeSnapshot(sid1, now, [(self.include.name, 0),]))
        self.assertTrue(sid1.exists())
        self.assertNotExists(sid1.path('leftover'))

    @patch('time.sleep') # speed up unittest
    def test_new_exists_continue(self, sleep):
        new_snapshot = snapshots.NewSnapshot(self.cfg)
        new_snapshot.makeDirs()
        with open(new_snapshot.path('leftover'), 'wt') as f:
            f.write('foo')
        new_snapshot.saveToContinue = True

        now = datetime.today() - timedelta(minutes = 6)
        sid1 = snapshots.SID(now, self.cfg)

        self.assertListEqual([True, False], self.sn.takeSnapshot(sid1, now, [(self.include.name, 0),]))
        self.assertTrue(sid1.exists())
        self.assertExists(sid1.path('leftover'))

    @patch('time.sleep')  # speed up unittest
    def test_fail_create_new_snapshot(self, sleep):
        with generic.mockPermissions(self.snapshotPath, 0o500):
            now = datetime.today()
            sid1 = snapshots.SID(now, self.cfg)

            self.assertListEqual([False, True], self.sn.takeSnapshot(sid1, now, [(self.include.name, 0),]))


@unittest.skipIf(
    not generic.LOCAL_SSH,
    'Skip as this test requires a local ssh server, '
    'public and private keys installed')
class TakeSSH(generic.SSHSnapshotTestCase, Take):
    def setUp(self):
        super().setUp()
        self.include = TemporaryDirectory()
        generic.create_test_files(self.include.name)

        #mount
        self.cfg.setCurrentHashId(mount.Mount(cfg = self.cfg).mount())

    def tearDown(self):
        #unmount
        mount.Mount(cfg = self.cfg).umount(self.cfg.current_hash_id)
        super().tearDown()

        self.include.cleanup()

    def remount(self):
        mount.Mount(cfg = self.cfg).umount(self.cfg.current_hash_id)
        mount.Mount(cfg = self.cfg).mount()

    def getInode(self, sid):
        return os.stat(os.path.join(self.snapshotPath, sid.sid, 'backup', self.include.name[1:], 'test')).st_ino



# SPDX-FileCopyrightText: © 2012-2022 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
import sys
import os
import time
import atexit
import signal
import config
import configfile
import tools
import daemon
import password_ipc
import logger
from exceptions import Timeout


class Password_Cache(daemon.Daemon):
    """
    Password_Cache get started on User login. It provides passwords for
    BIT cronjobs because keyring is not available when the User is not
    logged in. Does not start if there is no password to cache
    (e.g. no profile allows to cache).
    """
    PW_CACHE_VERSION = 3

    def __init__(self, cfg = None, *args, **kwargs):
        self.config = cfg
        if self.config is None:
            self.config = config.Config()
        cachePath = self.config.passwordCacheFolder()
        if not tools.mkdir(cachePath, 0o700):
            msg = 'Failed to create secure Password_Cache folder'
            logger.error(msg, self)
            raise PermissionError(msg)
        pid = self.config.passwordCachePid()
        super(Password_Cache, self).__init__(pid, umask = 0o077, *args, **kwargs)
        self.dbKeyring = {}
        self.dbUsr = {}
        self.fifo = password_ipc.FIFO(self.config.passwordCacheFifo())

        self.keyringSupported = tools.keyringSupported()

    def run(self):
        """
        wait for password request on FIFO and answer with password
        from self.db through FIFO.
        """
        info = configfile.ConfigFile()
        info.setIntValue('version', self.PW_CACHE_VERSION)
        info.save(self.config.passwordCacheInfo())
        os.chmod(self.config.passwordCacheInfo(), 0o600)

        logger.debug('Keyring supported: %s' %self.keyringSupported, self)

        tools.envSave(self.config.cronEnvFile())

        if not self.collectPasswords():
            logger.debug('Nothing to cache. Quit.', self)
            sys.exit(0)
        self.fifo.create()
        atexit.register(self.fifo.delfifo)
        signal.signal(signal.SIGHUP, self.reloadHandler)
        logger.debug('Start loop', self)

        while True:
            try:
                request = self.fifo.read()
                request = request.split('\n')[0]
                task, value = request.split(':', 1)

                if task == 'get_pw':
                    key = value

                    if key in list(self.dbKeyring.keys()):
                        answer = 'pw:' + self.dbKeyring[key]
                    elif key in list(self.dbUsr.keys()):
                        answer = 'pw:' + self.dbUsr[key]
                    else:
                        answer = 'none:'

                    self.fifo.write(answer, 5)

                elif task == 'set_pw':
                    key, value = value.split(':', 1)
                    self.dbUsr[key] = value

            except IOError as e:
                logger.error('Error in writing answer to FIFO: %s' % str(e), self)

            except KeyboardInterrupt:
                logger.debug('Quit.', self)
                break

            except Timeout:
                # That exception is thrown by tools.Alarm.handle() if the
                # timeout ends. That Alarm was set by FIFO.read().
                logger.error('FIFO timeout', self)

            except Exception as e:
                logger.error('ERROR: %s' % str(e), self)

    def reloadHandler(self, signum, frame):
        """
        reload passwords during runtime.
        """
        time.sleep(2)
        cfgPath = self.config._LOCAL_CONFIG_PATH
        del self.config
        self.config = config.Config(cfgPath)
        del self.dbKeyring
        self.dbKeyring = {}
        self.collectPasswords()

    def collectPasswords(self):
        """
        search all profiles in config and collect passwords from keyring.
        """
        run_daemon = False
        profiles = self.config.profiles()
        for profile_id in profiles:
            mode = self.config.snapshotsMode(profile_id)
            for pw_id in (1, 2):
                if self.config.modeNeedPassword(mode, pw_id):
                    if self.config.passwordUseCache(profile_id):
                        run_daemon = True
                        if self.config.passwordSave(profile_id) and self.keyringSupported:
                            service_name = self.config.keyringServiceName(profile_id, mode, pw_id)
                            user_name = self.config.keyringUserName(profile_id)

                            password = tools.password(service_name, user_name)
                            if password is None:
                                continue
                            self.dbKeyring['%s/%s' %(service_name, user_name)] = password
        return run_daemon

    def checkVersion(self):
        info = configfile.ConfigFile()
        info.load(self.config.passwordCacheInfo())
        if info.intValue('version') < self.PW_CACHE_VERSION:
            return False
        return True

    def cleanupHandler(self, signum, frame):
        self.fifo.delfifo()
        super(Password_Cache, self).cleanupHandler(signum, frame)


class Password:
    """Provide passwords for BIT either from keyring, Password_Cache or
    by asking user.
    """

    def __init__(self, cfg=None):
        self.config = cfg

        if self.config is None:
            self.config = config.Config()

        self.cache = Password_Cache(self.config)
        self.fifo = password_ipc.FIFO(self.config.passwordCacheFifo())
        self.db = {}

        self.keyringSupported = tools.keyringSupported()

    def password(self,
                 parent,
                 profile_id,
                 mode,
                 pw_id=1,
                 only_from_keyring=False,
                 refresh=False):
        """
        Based on profile settings return password from keyring,
        Password_Cache or by asking User.
        """
        if not self.config.modeNeedPassword(mode, pw_id):
            return ''

        service_name = self.config.keyringServiceName(profile_id, mode, pw_id)

        user_name = self.config.keyringUserName(profile_id)

        try:
            return self.db['%s/%s' % (service_name, user_name)]

        except KeyError:
            pass

        password = ''

        if (self.config.passwordUseCache(profile_id)
                and not only_from_keyring
                and not refresh):
            # From cache
            password = self.passwordFromCache(service_name, user_name)

            if password is not None:
                self.setPasswordDb(service_name, user_name, password)

                return password

        if self.config.passwordSave(profile_id) and not refresh:
            # From keyring
            password = self.passwordFromKeyring(service_name, user_name)

            if password is not None:
                self.setPasswordDb(service_name, user_name, password)

                return password

        if refresh or not only_from_keyring:
            # Ask user and write to cache
            password = self.passwordFromUser(parent, profile_id, mode, pw_id)

            if self.config.passwordUseCache(profile_id):
                self.setPasswordCache(service_name, user_name, password)

            self.setPasswordDb(service_name, user_name, password)

            return password

        return password

    def passwordFromKeyring(self, service_name, user_name):
        """
        get password from system keyring (seahorse). The keyring is only
        available if User is logged in.
        """
        if self.keyringSupported:
            try:
                return tools.password(service_name, user_name)
            except Exception:
                logger.error('get password from Keyring failed', self)
        return None

    def passwordFromCache(self, service_name, user_name):
        """
        get password from Password_Cache
        """
        if self.cache.status():
            self.cache.checkVersion()
            self.fifo.write('get_pw:%s/%s' %(service_name, user_name), timeout = 5)
            answer = self.fifo.read(timeout = 5)
            mode, pw = answer.split(':', 1)
            if mode == 'none':
                return None
            return pw
        else:
            return None

    def passwordFromUser(self,
                         parent,
                         profile_id=None,
                         mode=None,
                         pw_id=1,
                         prompt=None):
        """Ask user for password.

        Use terminal input or a dialog box if X is available. This does even
        work when run as cronjob and user is logged in.

        Args:
            parent: Parent of the ``QMessageDialog``.
            profile_id(str): Profile identifier.
            mode(str): Mode identifier (e.g. SSH (encrypted)).
            pwd_id(int): See :data:`config.SNAPSHOT_MODES` for details.
            prompt(str): Alternative string used as prompt.

        Return:
            str: The password.
        """
        # Default prompt
        if prompt is None:
            prompt = _('Enter password for {mode} profile "{profile}":') \
                .format(
                    profile=self.config.profileName(profile_id),
                    mode=self.config.SNAPSHOT_MODES[mode][pw_id+1])

        tools.registerBackintimePath('qt')

        x_server = tools.checkXServer()
        import_successful = False

        if x_server:
            try:
                import messagebox
                import_successful = True

            except ImportError:
                pass

        if not import_successful or not x_server:
            import getpass

            alarm = tools.Alarm()
            alarm.start(300)

            try:
                password = getpass.getpass(prompt + ' ')
                alarm.stop()

            except Timeout:
                password = ''

            return password

        # Use QDialog as graphical prompt
        password = messagebox.askPasswordDialog(
            parent=parent,
            title=self.config.APP_NAME,
            prompt=prompt,
            language_code=self.config.language(),
            timeout=300)

        return password

    def setPasswordDb(self, service_name, user_name, password):
        """
        internal Password cache. Prevent to ask password several times
        during runtime.
        """
        self.db['%s/%s' %(service_name, user_name)] = password

    def setPassword(self, password, profile_id, mode, pw_id):
        """
        store password to keyring and Password_Cache
        """
        if self.config.modeNeedPassword(mode, pw_id):
            service_name = self.config.keyringServiceName(profile_id, mode, pw_id)
            user_name = self.config.keyringUserName(profile_id)

            if self.config.passwordSave(profile_id):
                self.setPasswordKeyring(service_name, user_name, password)

            if self.config.passwordUseCache(profile_id):
                self.setPasswordCache(service_name, user_name, password)

            self.setPasswordDb(service_name, user_name, password)

    def setPasswordKeyring(self, service_name, user_name, password):
        return tools.setPassword(service_name, user_name, password)

    def setPasswordCache(self, service_name, user_name, password):
        if self.cache.status():
            self.cache.checkVersion()
            self.fifo.write('set_pw:%s/%s:%s' %(service_name, user_name, password), timeout = 5)



# SPDX-FileCopyrightText: © 2024 Christian BUHTZ <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See file/folder LICENSE or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""Module with an improved combo box widget."""
from typing import Any
from PyQt6.QtWidgets import QComboBox, QWidget


class BitComboBox(QComboBox):
    """Improved combo box.

    This widget can be filled (with content and data) just by a dictionary. It
    has the ability to select a specific entry based on its underlying
    `userData` instead of just the index.

    Example of a dictionary : ::

        # Values in the dictionary are the strings displayed in the combo box.
        # The keys are the underlying 'userData'.
        fill = {
            10: 'Hour',
            20: 'Day',
            30: 'Week',
            40: 'Month'
        }
        combo = BitComboBox(parent, fill)

    """

    def __init__(self, parent: QWidget, content_dict: dict):
        """
        Args:
            parent: The parent widget.
            content_dict: The dictionary values used to display entries in the
                combo box and the keys used as data.
        """
        super().__init__(parent=parent)

        self._content_dict = content_dict

        for data, entry in self._content_dict.items():
            self.addItem(entry, userData=data)

    @property
    def current_data(self) -> Any:
        """Data linked to the current selected entry."""
        return self.itemData(self.currentIndex())

    def select_by_data(self, data: Any):
        """Select an entry in the combo box by its underlying data."""
        for idx in range(self.count()):
            if self.itemData(idx) == data:
                self.setCurrentIndex(idx)
                return

        raise ValueError('Unable to select combo box entry because data not '
                         f'found in it. Data is: {data} (type: {type(data)})')



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.

# TODO Known open issues:
# - this script should get started and consider some cmd line arguments from BiT
#   (parsed via backintime.createParsers()) so that the same paths are used,
#   mainly "share-path" and "config" (path to the config file).
#   Otherwise e.g. unit tests or special user path settings may lead to
#   wrong status info in the systray icon!

import sys
import os
import pluginmanager
import tools
import logger
import gettext
import subprocess


_ = gettext.gettext


if not os.getenv('DISPLAY', ''):
    os.putenv('DISPLAY', ':0.0')


class SysTrayIconPlugin(pluginmanager.Plugin):
    def __init__(self):
        self.process = None
        self.snapshots = None

    def init(self, snapshots):
        self.snapshots = snapshots

        # Old implementation disabled:
        # Why can a systray icon only be shown on X11 (not wayland)?
        # Qt can handle wayland now!
        #    if not tools.checkXServer():
        #        return False

        # New implementation: Let Qt decide if a system tray icon can be shown.
        # See https://doc.qt.io/qt-5/qsystemtrayicon.html#details:
        # > To check whether a system tray is present on the user's desktop,
        # > call the QSystemTrayIcon::isSystemTrayAvailable() static function.
        #
        # This requires a QApplication instance (otherwise Qt causes a segfault)
        # which we don't have here so we create it to check if a window manager
        # ("GUI") is active at all (e.g. in headless installations it isn't).
        # See: https://forum.qt.io/topic/3852/issystemtrayavailable-always-crashes-segfault-on-ubuntu-10-10-desktop/6

        try:

            if tools.is_Qt_working(systray_required=True):
                logger.debug("System tray is available to show the BiT system tray icon")
                return True

        except Exception as e:
            logger.debug(f"Could not ask Qt if system tray is available: {repr(e)}")

        logger.debug("No system tray available to show the BiT system tray icon")
        return False

    def isGui(self):
        return True

    def processBegin(self):
        try:
            logger.debug("Trying to start systray icon sub process...")
            path = os.path.join(tools.backintimePath('qt'), 'qtsystrayicon.py')
            cmd = [sys.executable, path, self.snapshots.config.currentProfile()]
            if logger.DEBUG:
                cmd.append("--debug")  # HACK to propagate DEBUG logging level to sub process
            self.process = subprocess.Popen(cmd)
            # self.process = subprocess.Popen([sys.executable, path, self.snapshots.config.currentProfile()])
        except:
            pass

    def processEnd(self):
        if not self.process is None:
            try:
                # The "qtsystrayicon.py" app does terminate itself
                # once the snapshot has been taken so there is no need
                # to do anything here to stop it or clean-up anything.
                # self.process.terminate()
                return
            except:
                pass



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
# SPDX-FileCopyrightText: © 2008-2022 Taylor Raak
# SPDX-FileCopyrightText: © 2024 Christian BUHTZ <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
import os
import re
from PyQt6.QtWidgets import (QVBoxLayout,
                             QDialogButtonBox,
                             QPlainTextEdit,
                             QDialog)

import tools
import logger


class EditUserCallback(QDialog):
    def __init__(self, parent):
        super(EditUserCallback, self).__init__(parent)
        self.config = parent.config
        self.script = self.config.takeSnapshotUserCallback()

        import icon
        self.setWindowIcon(icon.SETTINGS_DIALOG)
        self.setWindowTitle(self.script)
        self.resize(800, 500)

        layout = QVBoxLayout(self)
        self.edit = QPlainTextEdit(self)

        try:
            with open(self.script, 'rt') as f:
                self.edit.setPlainText(f.read())

        except IOError:
            pass

        layout.addWidget(self.edit)

        btnBox = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel,
            parent=self)

        btnBox.accepted.connect(self.accept)
        btnBox.rejected.connect(self.reject)
        layout.addWidget(btnBox)

    def checkScript(self, script):
        m = re.match(r'^#!(/[\w/-]+)\n', script)

        if not m:
            logger.error(
                'user-callback script has no shebang (#!/bin/sh) line.')
            self.config.errorHandler(
                'user-callback script has no shebang (#!/bin/sh) line.')

            return False

        if not tools.checkCommand(m.group(1)):
            logger.error('Shebang in user-callback script is not executable.')
            self.config.errorHandle(
                'Shebang in user-callback script is not executable.')

            return False

        return True

    def accept(self):
        if not self.checkScript(self.edit.toPlainText()):
            return

        with open(self.script, 'wt') as f:
            f.write(self.edit.toPlainText())

        os.chmod(self.script, 0o755)

        super(EditUserCallback, self).accept()



# SPDX-FileCopyrightText: © 2024 Christian BUHTZ <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See file/folder LICENSE or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""Module with a widget combining a spinbox and a combobox."""
from typing import Any
from PyQt6.QtWidgets import QSpinBox, QWidget, QHBoxLayout
from manageprofiles.combobox import BitComboBox


class SpinBoxWithUnit(QWidget):
    """A combination of a `QspinBox` and `BitComboBox` (`QComboBox`).
    """

    def __init__(self,
                 parent: QWidget,
                 range_min_max: tuple[int, int],
                 content_dict: dict):
        """
        Args:
            parent: The parent widget.
            range_min_max: ...
            content_dict: The dictionary values used to display entries in the
                combo box and the keys used as data.
        """
        super().__init__(parent=parent)

        layout = QHBoxLayout(self)

        self._spin = QSpinBox(self)
        self._spin.setRange(*range_min_max)
        layout.addWidget(self._spin)

        self._combo = BitComboBox(self, content_dict)
        layout.addWidget(self._combo)

    @property
    def data_and_unit(self) -> tuple[int, Any]:
        """Data linked to the current selected entry."""
        return (self._spin.value(), self._combo.current_data)

    def select_unit(self, data: Any):
        """Select a unit entry in the combo box by its underlying data."""
        self._combo.select_by_data(data)

    def unit(self) -> Any:
        return self._combo.current_data

    def value(self) -> int:
        """Get value of spin box."""
        return self._spin.value()

    def set_value(self, val: int) -> None:
        """Set value of spin box."""
        self._spin.setValue(val)



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
import os
import sys
import unittest
import stat
from datetime import date, datetime
from test import generic
from unittest.mock import patch

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
import configfile
import snapshots
from snapshotlog import LogFilter, SnapshotLog


class TestSID(generic.SnapshotsTestCase):
    def test_new_object_with_valid_date(self):
        sid1 = snapshots.SID('20151219-010324-123', self.cfg)
        sid2 = snapshots.SID('20151219-010324', self.cfg)
        sid3 = snapshots.SID(datetime(2015, 12, 19, 1, 3, 24), self.cfg)
        sid4 = snapshots.SID(date(2015, 12, 19), self.cfg)

        self.assertEqual(sid1.sid,  '20151219-010324-123')
        self.assertEqual(sid2.sid,  '20151219-010324')
        self.assertRegex(sid3.sid, r'20151219-010324-\d{3}')
        self.assertRegex(sid4.sid, r'20151219-000000-\d{3}')

    def test_new_object_with_invalid_value(self):
        with self.assertRaises(ValueError):
            snapshots.SID('20151219-010324-1234', self.cfg)
        with self.assertRaises(ValueError):
            snapshots.SID('20151219-01032', self.cfg)
        with self.assertRaises(ValueError):
            snapshots.SID('2015121a-010324-1234', self.cfg)

    def test_new_object_with_invalid_type(self):
        with self.assertRaises(TypeError):
            snapshots.SID(123, self.cfg)

    def test_equal_sid(self):
        sid1a = snapshots.SID('20151219-010324-123', self.cfg)
        sid1b = snapshots.SID('20151219-010324-123', self.cfg)
        sid2  = snapshots.SID('20151219-020324-123', self.cfg)

        self.assertIsNot(sid1a, sid1b)
        self.assertTrue(sid1a == sid1b)
        self.assertTrue(sid1a == '20151219-010324-123')
        self.assertTrue(sid1a != sid2)
        self.assertTrue(sid1a != '20151219-020324-123')

    def test_sort_sids(self):
        root = snapshots.RootSnapshot(self.cfg)
        new  = snapshots.NewSnapshot(self.cfg)
        sid1 = snapshots.SID('20151219-010324-123', self.cfg)
        sid2 = snapshots.SID('20151219-020324-123', self.cfg)
        sid3 = snapshots.SID('20151219-030324-123', self.cfg)
        sid4 = snapshots.SID('20151219-040324-123', self.cfg)

        sids1 = [sid3, sid1, sid4, sid2]
        sids1.sort()
        self.assertEqual(sids1, [sid1, sid2, sid3, sid4])

        #RootSnapshot 'Now' should always stay on top
        sids2 = [sid3, sid1, root, sid4, sid2]
        sids2.sort()
        self.assertEqual(sids2, [sid1, sid2, sid3, sid4, root])

        sids2.sort(reverse = True)
        self.assertEqual(sids2, [root, sid4, sid3, sid2, sid1])

        #new_snapshot should always be the last
        sids3 = [sid3, sid1, new, sid4, sid2]
        sids3.sort()
        self.assertEqual(sids3, [sid1, sid2, sid3, sid4, new])

        sids3.sort(reverse = True)
        self.assertEqual(sids3, [new, sid4, sid3, sid2, sid1])

    def test_hash(self):
        sid1a = snapshots.SID('20151219-010324-123', self.cfg)
        sid1b = snapshots.SID('20151219-010324-123', self.cfg)
        sid2  = snapshots.SID('20151219-020324-123', self.cfg)

        self.assertEqual(sid1a.__hash__(), sid1b.__hash__())
        self.assertNotEqual(sid1a.__hash__(), sid2.__hash__())

        s = set()
        s.add(sid1a)
        self.assertEqual(len(s), 1)
        s.add(sid2)
        self.assertEqual(len(s), 2)
        s.add(sid1b)
        self.assertEqual(len(s), 2)

    def test_split(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)

        self.assertTupleEqual(sid.split(), (2015, 12, 19, 1, 3, 24))

    def test_displayID(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)

        self.assertEqual(sid.displayID, '2015-12-19 01:03:24')

    def test_displayName(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        os.makedirs(os.path.join(self.snapshotPath, '20151219-010324-123'))
        with open(sid.path('name'), 'wt') as f:
            f.write('foo')

        self.assertEqual(sid.displayName, '2015-12-19 01:03:24 - foo')

        with open(sid.path('failed'), 'wt') as f:
            pass

        self.assertRegex(sid.displayName, r'2015-12-19 01:03:24 - foo (.+?)')

    def test_withoutTag(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)

        self.assertEqual(sid.withoutTag, '20151219-010324')

    def test_tag(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)

        self.assertEqual(sid.tag, '123')

    def test_path(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)

        self.assertEqual(sid.path(),
                         os.path.join(self.snapshotPath, '20151219-010324-123'))
        self.assertEqual(sid.path('foo', 'bar', 'baz'),
                         os.path.join(self.snapshotPath,
                                      '20151219-010324-123',
                                      'foo', 'bar', 'baz'))
        self.assertEqual(sid.pathBackup(),
                         os.path.join(self.snapshotPath,
                                      '20151219-010324-123',
                                      'backup'))
        self.assertEqual(sid.pathBackup('/foo'),
                         os.path.join(self.snapshotPath,
                                      '20151219-010324-123',
                                      'backup', 'foo'))

    def test_makeDirs(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        self.assertTrue(sid.makeDirs())
        self.assertTrue(os.path.isdir(os.path.join(self.snapshotPath,
                                                   '20151219-010324-123',
                                                   'backup')))

        self.assertTrue(sid.makeDirs('foo', 'bar', 'baz'))
        self.assertTrue(os.path.isdir(os.path.join(self.snapshotPath,
                                                   '20151219-010324-123',
                                                   'backup',
                                                   'foo', 'bar', 'baz')))

    def test_exists(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        self.assertFalse(sid.exists())

        os.makedirs(os.path.join(self.snapshotPath, '20151219-010324-123'))
        self.assertFalse(sid.exists())

        os.makedirs(os.path.join(self.snapshotPath,
                                 '20151219-010324-123',
                                 'backup'))
        self.assertTrue(sid.exists())

    def test_isExistingPathInsideSnapshotFolder(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        backupPath = os.path.join(self.snapshotPath,
                                  '20151219-010324-123',
                                  'backup')
        os.makedirs(os.path.join(backupPath, 'foo'))

        #test existing file and non existing file
        self.assertTrue(sid.isExistingPathInsideSnapshotFolder('/foo'))
        self.assertFalse(sid.isExistingPathInsideSnapshotFolder('/tmp'))

        #test valid absolute symlink inside snapshot
        os.symlink(os.path.join(backupPath, 'foo'),
                   os.path.join(backupPath, 'bar'))
        self.assertIsLink(backupPath, 'bar')
        self.assertTrue(sid.isExistingPathInsideSnapshotFolder('/bar'))

        #test valid relative symlink inside snapshot
        os.symlink('./foo',
                   os.path.join(backupPath, 'baz'))
        self.assertIsLink(backupPath, 'baz')
        self.assertTrue(sid.isExistingPathInsideSnapshotFolder('/baz'))

        #test invalid symlink
        os.symlink(os.path.join(backupPath, 'asdf'),
                   os.path.join(backupPath, 'qwer'))
        self.assertIsLink(backupPath, 'qwer')
        self.assertFalse(sid.isExistingPathInsideSnapshotFolder('/qwer'))

        #test valid symlink outside snapshot
        os.symlink('/tmp',
                   os.path.join(backupPath, 'bla'))
        self.assertIsLink(backupPath, 'bla')
        self.assertFalse(sid.isExistingPathInsideSnapshotFolder('/bla'))

    def test_name(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        os.makedirs(os.path.join(self.snapshotPath, '20151219-010324-123'))
        self.assertEqual(sid.name, '')

        sid.name = 'foo'
        with open(sid.path('name'), 'rt') as f:
            self.assertEqual(f.read(), 'foo')

        self.assertEqual(sid.name, 'foo')

    def test_lastChecked(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        os.makedirs(os.path.join(self.snapshotPath, '20151219-010324-123'))
        infoFile = os.path.join(self.snapshotPath, '20151219-010324-123', 'info')

        #no info file
        self.assertEqual(sid.lastChecked, '2015-12-19 01:03:24')

        #set time manually to 2015-12-19 02:03:24
        with open(infoFile, 'wt'):
            pass
        d = datetime(2015, 12, 19, 2, 3, 24)
        os.utime(infoFile, (d.timestamp(), d.timestamp()))
        self.assertEqual(sid.lastChecked, '2015-12-19 02:03:24')

        #setLastChecked and check if it matches current date
        sid.setLastChecked()
        now = datetime.now()
        self.assertRegex(sid.lastChecked, now.strftime(r'%Y-%m-%d %H:%M:\d{2}'))

    def test_failed(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        snapshotPath = os.path.join(self.snapshotPath, '20151219-010324-123')
        failedPath   = os.path.join(snapshotPath, sid.FAILED)
        os.makedirs(snapshotPath)

        self.assertNotExists(failedPath)
        self.assertFalse(sid.failed)
        sid.failed = True
        self.assertExists(failedPath)
        self.assertTrue(sid.failed)

        sid.failed = False
        self.assertNotExists(failedPath)
        self.assertFalse(sid.failed)

    def test_info(self):
        sid1 = snapshots.SID('20151219-010324-123', self.cfg)
        os.makedirs(os.path.join(self.snapshotPath, '20151219-010324-123'))
        infoFile = os.path.join(self.snapshotPath, '20151219-010324-123', 'info')

        i1 = configfile.ConfigFile()
        i1.setStrValue('foo', 'bar')
        sid1.info = i1

        #test if file exist and has correct content
        self.assertIsFile(infoFile)
        with open(infoFile, 'rt') as f:
            self.assertEqual(f.read(), 'foo=bar\n')

        #new sid instance and test if correct value is returned
        sid2 = snapshots.SID('20151219-010324-123', self.cfg)
        i2 = sid2.info
        self.assertEqual(i2.strValue('foo', 'default'), 'bar')

    def test_fileInfo(self):
        sid1 = snapshots.SID('20151219-010324-123', self.cfg)
        os.makedirs(os.path.join(self.snapshotPath, '20151219-010324-123'))
        infoFile = os.path.join(self.snapshotPath,
                                '20151219-010324-123',
                                'fileinfo.bz2')

        d = snapshots.FileInfoDict()
        d[b'/tmp']     = (123, b'foo', b'bar')
        d[b'/tmp/foo'] = (456, b'asdf', b'qwer')
        sid1.fileInfo = d

        self.assertIsFile(infoFile)

        #load fileInfo in a new snapshot
        sid2 = snapshots.SID('20151219-010324-123', self.cfg)
        self.assertDictEqual(sid2.fileInfo, d)

    @patch('logger.error')
    def test_fileInfoErrorRead(self, mock_logger):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        os.makedirs(os.path.join(self.snapshotPath, '20151219-010324-123'))
        infoFile = sid.path(sid.FILEINFO)
        # remove all permissions from file
        with open(infoFile, 'wt'):
            pass

        with generic.mockPermissions(infoFile):
            self.assertEqual(sid.fileInfo, snapshots.FileInfoDict())
            self.assertTrue(mock_logger.called)

    @patch('logger.error')
    def test_fileInfoErrorWrite(self, mock_logger):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        os.makedirs(os.path.join(self.snapshotPath, '20151219-010324-123'))
        infoFile = sid.path(sid.FILEINFO)
        # remove all permissions from file
        with open(infoFile, 'wt'):
            pass

        with generic.mockPermissions(infoFile):
            d = snapshots.FileInfoDict()
            d[b'/tmp']     = (123, b'foo', b'bar')
            d[b'/tmp/foo'] = (456, b'asdf', b'qwer')
            sid.fileInfo = d
            self.assertTrue(mock_logger.called)

    def test_log(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        os.makedirs(os.path.join(self.snapshotPath, '20151219-010324-123'))
        logFile = os.path.join(self.snapshotPath,
                               '20151219-010324-123',
                               'takesnapshot.log.bz2')

        #no log available
        self.assertRegex('\n'.join(sid.log()), r'Failed to get snapshot log from.*')

        sid.setLog('foo bar\nbaz')
        self.assertIsFile(logFile)

        self.assertEqual('\n'.join(sid.log()), 'foo bar\nbaz')

    def test_log_filter(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        os.makedirs(os.path.join(self.snapshotPath, '20151219-010324-123'))
        logFile = os.path.join(self.snapshotPath,
                               '20151219-010324-123',
                               'takesnapshot.log.bz2')

        sid.setLog('foo bar\n[I] 123\n[C] baz\n[E] bla')
        self.assertIsFile(logFile)

        self.assertEqual('\n'.join(sid.log(mode = LogFilter.CHANGES)), 'foo bar\n[C] baz')

    def test_setLog_binary(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        os.makedirs(os.path.join(self.snapshotPath, '20151219-010324-123'))
        logFile = os.path.join(self.snapshotPath,
                               '20151219-010324-123',
                               'takesnapshot.log.bz2')

        sid.setLog(b'foo bar\nbaz')
        self.assertIsFile(logFile)

        self.assertEqual('\n'.join(sid.log()), 'foo bar\nbaz')

    def test_makeWritable(self):
        sid = snapshots.SID('20151219-010324-123', self.cfg)
        sidPath = os.path.join(self.snapshotPath,   '20151219-010324-123')
        os.makedirs(sidPath)
        testFile = os.path.join(sidPath, 'test')

        #make only read and explorable
        os.chmod(sidPath, stat.S_IRUSR | stat.S_IXUSR)
        with self.assertRaises(PermissionError):
            with open(testFile, 'wt') as f:
                f.write('foo')

        sid.makeWritable()

        self.assertEqual(os.stat(sidPath).st_mode & stat.S_IWUSR, stat.S_IWUSR)
        try:
            with open(testFile, 'wt') as f:
                f.write('foo')
        except PermissionError:
            msg = 'writing to {} raised PermissionError unexpectedly!'
            self.fail(msg.format(testFile))


class TestNewSnapshot(generic.SnapshotsTestCase):
    def test_create_new(self):
        new = snapshots.NewSnapshot(self.cfg)
        self.assertFalse(new.exists())

        self.assertTrue(new.makeDirs())
        self.assertTrue(new.exists())
        self.assertTrue(os.path.isdir(os.path.join(self.snapshotPath,
                                                   new.NEWSNAPSHOT,
                                                   'backup')))

    def test_saveToContinue(self):
        new = snapshots.NewSnapshot(self.cfg)
        snapshotPath = os.path.join(self.snapshotPath, new.NEWSNAPSHOT)
        saveToContinuePath = os.path.join(snapshotPath, new.SAVETOCONTINUE)
        os.makedirs(snapshotPath)

        self.assertNotExists(saveToContinuePath)
        self.assertFalse(new.saveToContinue)

        new.saveToContinue = True
        self.assertExists(saveToContinuePath)
        self.assertTrue(new.saveToContinue)

        new.saveToContinue = False
        self.assertNotExists(saveToContinuePath)
        self.assertFalse(new.saveToContinue)

    def test_hasChanges(self):
        now = datetime(2016, 7, 10, 16, 24, 17)

        new = snapshots.NewSnapshot(self.cfg)
        new.makeDirs()

        log = SnapshotLog(self.cfg)
        log.new(now)
        self.assertFalse(new.hasChanges)
        log.append('[I] foo', log.ALL)
        log.append('[E] bar', log.ALL)
        log.flush()
        self.assertFalse(new.hasChanges)
        log.append('[C] baz', log.ALL)
        log.flush()
        self.assertTrue(new.hasChanges)


class TestRootSnapshot(generic.SnapshotsTestCase):
    #TODO: add test with 'sid.path(use_mode=['ssh_encfs'])'
    def test_create(self):
        sid = snapshots.RootSnapshot(self.cfg)
        self.assertTrue(sid.isRoot)
        self.assertEqual(sid.sid, '/')
        self.assertEqual(sid.name, 'Now')

    def test_path(self):
        sid = snapshots.RootSnapshot(self.cfg)
        self.assertEqual(sid.path(), '/')
        self.assertEqual(sid.path('foo', 'bar'), '/foo/bar')


class TestIterSnapshots(generic.SnapshotsTestCase):
    def setUp(self):
        super(TestIterSnapshots, self).setUp()

        for i in ('20151219-010324-123',
                  '20151219-020324-123',
                  '20151219-030324-123',
                  '20151219-040324-123'):
            os.makedirs(os.path.join(self.snapshotPath, i, 'backup'))

    def test_list_valid(self):
        l1 = snapshots.listSnapshots(self.cfg)
        self.assertListEqual(l1, ['20151219-040324-123',
                                  '20151219-030324-123',
                                  '20151219-020324-123',
                                  '20151219-010324-123'])
        self.assertIsInstance(l1[0], snapshots.SID)

    def test_list_new_snapshot(self):
        os.makedirs(os.path.join(self.snapshotPath, 'new_snapshot', 'backup'))
        l2 = snapshots.listSnapshots(self.cfg, includeNewSnapshot = True)
        self.assertListEqual(l2, ['new_snapshot',
                                  '20151219-040324-123',
                                  '20151219-030324-123',
                                  '20151219-020324-123',
                                  '20151219-010324-123'])
        self.assertIsInstance(l2[0], snapshots.NewSnapshot)
        self.assertIsInstance(l2[-1], snapshots.SID)

    def test_list_snapshot_without_backup(self):
        #new snapshot without backup folder shouldn't be added
        os.makedirs(os.path.join(self.snapshotPath, '20151219-050324-123'))
        l3 = snapshots.listSnapshots(self.cfg)
        self.assertListEqual(l3, ['20151219-040324-123',
                                  '20151219-030324-123',
                                  '20151219-020324-123',
                                  '20151219-010324-123'])

    def test_list_invalid_snapshot(self):
        #invalid snapshot shouldn't be added
        os.makedirs(os.path.join(self.snapshotPath,
                                 '20151219-000324-abc',
                                 'backup'))
        l4 = snapshots.listSnapshots(self.cfg)
        self.assertListEqual(l4, ['20151219-040324-123',
                                  '20151219-030324-123',
                                  '20151219-020324-123',
                                  '20151219-010324-123'])

    def test_list_without_new_snapshot(self):
        os.makedirs(os.path.join(self.snapshotPath, 'new_snapshot', 'backup'))
        l5 = snapshots.listSnapshots(self.cfg, includeNewSnapshot = False)
        self.assertListEqual(l5, ['20151219-040324-123',
                                  '20151219-030324-123',
                                  '20151219-020324-123',
                                  '20151219-010324-123'])

    def test_list_symlink_last_snapshot(self):
        os.symlink('./20151219-040324-123',
                   os.path.join(self.snapshotPath, 'last_snapshot'))
        l6 = snapshots.listSnapshots(self.cfg)
        self.assertListEqual(l6, ['20151219-040324-123',
                                  '20151219-030324-123',
                                  '20151219-020324-123',
                                  '20151219-010324-123'])

    def test_list_not_reverse(self):
        os.makedirs(os.path.join(self.snapshotPath, 'new_snapshot', 'backup'))
        l7 = snapshots.listSnapshots(self.cfg,
                                     includeNewSnapshot = True,
                                     reverse = False)
        self.assertListEqual(l7, ['20151219-010324-123',
                                  '20151219-020324-123',
                                  '20151219-030324-123',
                                  '20151219-040324-123',
                                  'new_snapshot'])
        self.assertIsInstance(l7[0], snapshots.SID)
        self.assertIsInstance(l7[-1], snapshots.NewSnapshot)

    def test_iter_snapshots(self):
        for i, sid in enumerate(snapshots.iterSnapshots(self.cfg)):
            self.assertIn(sid, ['20151219-040324-123',
                                '20151219-030324-123',
                                '20151219-020324-123',
                                '20151219-010324-123'])
            self.assertIsInstance(sid, snapshots.SID)
        self.assertEqual(i, 3)

    def test_lastSnapshot(self):
        self.assertEqual(snapshots.lastSnapshot(self.cfg),
                         '20151219-040324-123')


class TestIterSnapshotsNonexistingSnapshotPath(generic.TestCaseSnapshotPath):
    def test_iterSnapshots(self):
        for _ in snapshots.iterSnapshots(self.cfg):
            self.fail('got unexpected snapshot')

    def test_listSnapshots(self):
        self.assertEqual(snapshots.listSnapshots(self.cfg), [])

    def test_lastSnapshots(self):
        self.assertIsNone(snapshots.lastSnapshot(self.cfg))


if __name__ == '__main__':
    unittest.main()



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
# SPDX-FileCopyrightText: © 2024 Christian BUHTZ <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In time" which is released under GNU
# General Public License v2 (GPLv2). See file/folder LICENSE or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""
Basic functions for handling Cron, Crontab, and other scheduling-related
features.
"""
import subprocess
import logger

_MARKER = '#Back In Time system entry, this will be edited by the gui:'
"""The string is used in crontab file to mark entries as owned by Back
In Time. **WARNING**: Don't modify that string in code because it is used
as match target while parsing the crontab file. See
:func:`remove_bit_from_crontab()` for details.
"""


def _determine_crontab_command() -> str:
    """Return the name of one of the supported crontab commands if available.

    Returns:
        (str): The command name. Usually "crontab" or "fcrontab".

    Raises:
        RuntimeError: If none of the supported commands available.
    """
    to_check_commands = ['crontab', 'fcrontab']
    for cmd in to_check_commands:
        proc = subprocess.run(
            ['which', cmd],
            stdout=subprocess.PIPE,
            check=False
        )
        if proc.returncode == 0:
            return cmd

    # syslog is not yet initialized
    logger.openlog()
    msg = 'Command ' + ' and '.join(to_check_commands) + ' not found.'
    logger.critical(msg)

    raise RuntimeError(msg)


CRONTAB_COMMAND = _determine_crontab_command()


def read_crontab():
    """Read current users crontab.

    On errors an empty list is returned.

    Returns:
        list: Crontab lines.
    """
    proc = subprocess.run(
        [CRONTAB_COMMAND, '-l'],
        check=False,
        capture_output=True,
        text=True)

    # Error?
    if proc.returncode != 0:
        # Ignore missing crontabs
        if not proc.stderr.startswith('no crontab for'):
            logger.error(f'Failed to get content via "{CRONTAB_COMMAND}". '
                         f'Return code of {proc.args} was {proc.returncode}. '
                         f'Stderr: {proc.stderr}')

        return []

    content = proc.stdout.split('\n')

    # Remove empty lines from the end
    try:
        while content[-1] == '':
            content = content[:-1]
    except IndexError:
        pass

    # Fixes issue #1181 (line count of empty crontab was 1 instead of 0)
    if content == ['']:
        content = []

    return content


def write_crontab(lines):
    """Write users crontab.

    This will overwrite the whole users crontab. So to keep the old crontab
    and only add new entries you need to read it first with
    :py:func:`tools.readCrontab`, append new entries to the list and write
    it back.

    Args:
        lines (list, tuple): Lines that should be written to crontab.

    Returns:
        bool: ``True`` if successful otherwise ``False``.

    """
    content = '\n'.join(lines)

    # Crontab needs to end with a newline
    if not content.endswith('\n'):
        content += '\n'

    # Pipe the content (via echo over stdout) to crontab's stdin
    with subprocess.Popen(['echo', content], stdout=subprocess.PIPE) as echo:

        try:
            subprocess.run(
                [CRONTAB_COMMAND, '-'],
                stdin=echo.stdout,
                check=True,
                capture_output=True,
                text=True
            )

        except subprocess.CalledProcessError as err:
            logger.error(
                f'Failed to write crontab lines with "{CRONTAB_COMMAND}". '
                f'Return code was {err.returncode}. '
                f'Error was:\n{err.stderr}')
            return False

    return True


def remove_bit_from_crontab(crontab):
    """Remove crontab entries related to backintime and having a marker line in
    the line before.

    Args:
        lines(list): List of crontab lines.
    """
    # Indices of lines containing the marker
    marker_indexes = list(filter(
        lambda idx: _MARKER in crontab[idx],
        range(len(crontab))
    ))

    # Check if there is a valid BIT entry after the marker lines
    for idx in marker_indexes[:]:
        try:
            if 'backintime' in crontab[idx+1]:
                continue
        except IndexError:
            pass

        # Remove the current index because the following line is not valid
        marker_indexes.remove(marker_indexes.index(idx))

    modified_crontab = crontab[:]

    # Remove the marker comment line and the following backintime line
    for idx in reversed(marker_indexes):
        del modified_crontab[idx:idx+2]

    return modified_crontab


def append_bit_to_crontab(crontab, bit_lines):
    """Add new entries to existing crontab content.

    Args:
        crontab(list): A list of strings as crontab lines.
        bit_lines(list): A list of strings as new crontab lines.

    Returns:
        list: The new crontab lines.
    """
    for line in bit_lines:
        crontab.append(_MARKER)
        crontab.append(line)

    return crontab


def is_cron_running():
    """Validate if a cron instance is running.

    The output of ``ps`` is searched (case-insensitive) via ``grep`` for the
    string ``cron``.

    Returns:
        bool: The answer.
    """

    with subprocess.Popen(['ps', '-eo', 'comm'], stdout=subprocess.PIPE) as ps:
        try:
            subprocess.run(
                ['grep', '--ignore-case', 'cron'],
                stdin=ps.stdout,
                stdout=subprocess.PIPE,
                check=True
            )
        except subprocess.CalledProcessError:
            return False

    return True



# SPDX-FileCopyrightText: © 2010 paul <paul@woland>
# SPDX-FileCopyrightText: © 2010 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See file/folder LICENSE or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""Module about UniquenessSet"""
import os
import logger
from tools import md5sum


class UniquenessSet:
    """Check for uniqueness or equality of file(s).

    Uniqueness means that there are no duplicates. Equality means that two
    elements have the same value or properties.

    Equality is checked only if a file is specified with ``equal_to`` otherwise
    uniqueness will be checked. The latter is based on previous file
    checks. The class store hashes of each file checked via ``check()`` or
    ``checkUnique()``.

    Dev note (buhtz, 2024-10): The class is used in SnapshotsDialog and only in
    the specific/rare case when single files are compared. The class does two
    different things implicit. Not sure if this is a good solution. My
    recommendation is to move that code/feature into 'qt' into or near the
    `SnapshotsDialog` class.
    """

    def __init__(self, deep_check=False, follow_symlink=False, equal_to=''):
        """
        Args:
            deep_check (bool): If ``True`` use deep check which will compare
                files md5sums if they are of same size but no
                hardlinks (don't have the same inode).
                If ``False`` use files size and mtime
            follow_symlink (bool): Check symlinks target instead of the
                link itself (default: ``False``).
            equal_to (str): Full path to file. If not empty only return
                equal files to the given path instead of unique
                files. (default: ``''``)
        """
        self.deep_check = deep_check
        self.follow_sym = follow_symlink

        # if not self._uniq_dict[size] -> size already checked with md5sum
        self._uniq_dict = {}

        # if (size, inode) in self._size_inode -> path is a hlink
        self._size_inode = set()

        self.equal_to = equal_to

        if equal_to:
            st = os.stat(equal_to)

            if self.deep_check:
                self.reference = (st.st_size, md5sum(equal_to))

            else:
                self.reference = (st.st_size, int(st.st_mtime))

    def check(self, input_path):
        """Check file ``input_path`` for either uniqueness or equality
        (depending on ``equal_to`` from constructor).

        Args:
            input_path (str):   full path to file

        Returns:
            bool: ``True`` if file is unique and ``equal_to``
                is empty. Or ``True`` if file is equal to file in
                ``equal_to``
        """
        path = input_path

        # follow symlinks ?
        if self.follow_sym and os.path.islink(input_path):
            path = os.readlink(input_path)

        if self.equal_to:
            return self.checkEqual(path)

        else:
            return self.checkUnique(path)

    def checkUnique(self, path):
        """Check file ``path`` for uniqueness and store a unique key for
        ``path``.

        This check is performed if `equal_to` is empty. By default
        (``deep_check is False``) the uniqueness is based on file size and
        mtime. If ``deep_check is True`` the uniqueness is based on inode
        number and the files size (or md5sum).

        Args:
            path (str): Full path to file.

        Returns:
            bool: ``True`` if file is unique.

        """
        logger.debug(f'{path=}')

        if self.deep_check:
            stat = os.stat(path)
            size, inode = stat.st_size, stat.st_ino

            # Hardlink?
            if (size, inode) in self._size_inode:
                logger.debug(
                    "[deep test]: skip, it's a duplicate (size, inode)", self)
                return False

            self._size_inode.add((size, inode))

            if size not in self._uniq_dict:
                # first item of that size
                unique_key = size
                logger.debug("[deep test]: store current size?", self)

            else:
                prev = self._uniq_dict[size]
                if prev:
                    # store md5sum instead of previously stored size
                    md5sum_prev = md5sum(prev)
                    self._uniq_dict[md5sum_prev] = prev
                    # remove the entry with that size
                    self._uniq_dict[size] = None
                    logger.debug(
                        "[deep test]: size duplicate, remove the size, store "
                        "prev md5sum", self)
                unique_key = md5sum(path)
                logger.debug("[deep test]: store current md5sum?", self)

        else:
            # store a tuple of (size, modification time)
            obj = os.stat(path)
            unique_key = (obj.st_size, int(obj.st_mtime))

        # store if not already present, then return True
        if unique_key not in self._uniq_dict:
            logger.debug(" >> ok, store!", self)
            self._uniq_dict[unique_key] = path

            return True

        logger.debug(" >> skip (it's a duplicate)", self)

        return False

    def checkEqual(self, path):
        """Check if ``path`` is equal to the file specified by ``equal_to``.

        Args:
            path (str): Full path to file.

        Returns:
            bool: ``True`` if file is equal.
        """
        st = os.stat(path)

        if self.deep_check:

            if self.reference[0] == st.st_size:
                return self.reference[1] == md5sum(path)

            return False

        else:
            return self.reference == (st.st_size, int(st.st_mtime))



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
# SPDX-FileCopyrightText: © 2008-2022 Taylor Raak
# SPDX-FileCopyrightText: © 2024 Christian BUHTZ <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See file/folder LICENSE or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""A widget to setup an SSH proxy."""
import getpass
from PyQt6.QtWidgets import (QVBoxLayout,
                             QHBoxLayout,
                             QWidget,
                             QLabel,
                             QLineEdit,
                             QCheckBox)
from PyQt6.QtCore import Qt

import qttools


class SshProxyWidget(QWidget):
    """Used in SSH snapshot profiles on the General tab.

    Dev note by buhtz (2024-04): Just a quick n dirty solution until the
    re-design and re-factoring of the whole dialog.
    """

    def __init__(self, parent, host, port, user):
        super().__init__(parent)

        if host == '':
            port = ''
            user = ''

        elif isinstance(port, int):
            port = str(port)

        vlayout = QVBoxLayout(self)
        # zero margins
        vlayout.setContentsMargins(0, 0, 0, 0)

        self._checkbox = QCheckBox(_('SSH Proxy'), self)
        vlayout.addWidget(self._checkbox)
        self._checkbox.stateChanged.connect(self._slot_checkbox_changed)

        hlayout = QHBoxLayout()
        vlayout.addLayout(hlayout)

        hlayout.addWidget(QLabel(_('Host:'), self))
        self.host_edit = QLineEdit(host, self)
        hlayout.addWidget(self.host_edit)

        hlayout.addWidget(QLabel(_('Port:'), self))
        self.port_edit = QLineEdit(port, self)
        hlayout.addWidget(self.port_edit)

        hlayout.addWidget(QLabel(_('User:'), self))
        self.user_edit = QLineEdit(user, self)
        hlayout.addWidget(self.user_edit)

        if host == '':
            self._disable()

        qttools.set_wrapped_tooltip(
            self,
            _('Connect to the target host via this proxy (also known as a '
              'jump host). See "-J" in the "ssh" command documentation or '
              '"ProxyJump" in "ssh_config" man page for details.')
        )

    def _slot_checkbox_changed(self, state):
        if Qt.CheckState(state) == Qt.CheckState.Checked:
            self._enable()
        else:
            self._disable()

    def _set_default(self):
        """Set GUI elements back to default."""
        self.host_edit.setText('')
        self.port_edit.setText('22')
        self.user_edit.setText(getpass.getuser())

    def _disable(self):
        self._set_default()
        self._enable(False)

    def _enable(self, enable=True):
        # QEdit and QLabel's
        lay = self.layout().itemAt(1)
        for idx in range(lay.count()):
            lay.itemAt(idx).widget().setEnabled(enable)

    def values(self) -> dict[str, str, str]:
        """The widgets values as a dict.

        Returns:
            dict: A 3-item dict with keys "host", "port" and "user".
        """
        if self._checkbox.isChecked():
            return {
                'host': self.host_edit.text(),
                'port': self.port_edit.text(),
                'user': self.user_edit.text(),
            }

        return {
            'host': '',
            'port': '',
            'user': '',
        }



# SPDX-FileCopyrightText: © 2012-2022 Germar Reitze
# SPDX-FileCopyrightText: © 2012-2022 Taylor Raack
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""The mount API.

    The high-level mount API is :py:class:`Mount` and handles mount,
    umount, remount and checks for *Back In Time*. The low-level mount API
    is :py:class:`MountControl`. The latter can be used to create own
    mounting services via subclassing it. See the following example.

    Example:

        See this template to create your own mounting service by inheriting
        from :py:class:`MountControl`. All you need to do is:

        - Add your settings in ``qt/settingsdialog.py``.
        - Add settings in ``common/config.py``.
        - Use the following template class ``MountDummy``, rename and modify
          it to your needs.
        - Please use ``self.currentMountpoint`` as your local mountpoint.
        - Your class should inherit from :py:class:`mount.MountControl`.

        As real usage example see the two classes :py:class:`sshtools.SSH` and
        :py:class:`encfstools.EncFS_mount`.

    This is the template: ::

        class MountDummy(mount.MountControl):
            def __init__(self, *args, **kwargs):
                super(MountDummy, self).__init__(*args, **kwargs)

                self.all_kwargs = {}

                # First we need to map the settings.
                # If <arg> is in kwargs (e.g. if this class is called with
                # dummytools.Dummy(<arg> = <value>) this will map self.<arg> to
                # kwargs[<arg>]; else self.<arg> = <default> from config
                # e.g. self.setattrKwargs(<arg>, <default>, **kwargs)
                self.setattrKwargs(
                    'user', self.config.get_dummy_user(self.profile_id), **kwargs)
                self.setattrKwargs(
                    'host', self.config.get_dummy_host(self.profile_id), **kwargs)
                self.setattrKwargs(
                    'port', self.config.get_dummy_port(self.profile_id), **kwargs)
                self.setattrKwargs(
                    'password',
                    self.config.password(self.parent, self.profile_id),
                    store = False, **kwargs)

                self.setDefaultArgs()

                # If self.currentMountpoint is not the remote snapshot path
                # you can specify a subfolder of self.currentMountpoint for
                # the symlink
                self.symlink_subfolder = None

                self.mountproc = 'dummy'
                self.log_command = '%s: %s@%s' % (self.mode, self.user, self.host)

            def _mount(self):
                # Mount the service
                # Implement your mountprocess here.
                pass

            def _umount(self):
                # Umount the service
                # Implement your unmountprocess here.
                pass

            def preMountCheck(self, first_run = False):
                # Check what ever conditions must be given for the mount to be
                # done successful.
                # Raise MountException('Error description') if service can not mount
                # return True if everything is okay
                # all pre|post_[u]mount_check can also be used to prepare
                # things or clean up
                return True

            def postMountCheck(self):
                # Check if mount was successful
                # Raise MountException('Error description') if not
                return True

            def preUmountCheck(self):
                # Check if service is safe to umount
                # Raise MountException('Error description') if not
                return True

            def postUmountCheck(self):
                # Check if umount successful
                # Raise MountException('Error description') if not
                return True

"""
import getpass
import json
import os
import subprocess
from time import sleep
from zlib import crc32
from pathlib import Path
import config
import logger
import password
import tools
from exceptions import HashCollision, MountException


class Mount:
    """
    This is the high-level mount API. This will handle mount, umount, remount
    and checks on the low-level :py:class:`MountControl` subclass backends for
    BackInTime.

    If ``cfg`` is ``None`` this will load the default config. If ``profile_id``
    is ``None`` it will use
    :py:func:`configfile.ConfigFileWithProfiles.currentProfile`.

    If the current profile uses Password-Cache and the Password-Cache is not
    running this will try to start it.

    Args:
        cfg (config.Config): Current config.
        profile_id (str): Profile ID to be used.
        tmp_mount (bool): If ``True`` mount to a temporary destination.
        parent (QWidget): Parent widget for QDialogs or ``None`` if there
            is no parent.
    """

    def __init__(self,
                 cfg=None,
                 profile_id=None,
                 tmp_mount=False,
                 parent=None):
        self.config = cfg or config.Config()
        self.profile_id = profile_id or self.config.currentProfile()
        self.tmp_mount = tmp_mount
        self.parent = parent

        if self.config.passwordUseCache(self.profile_id):
            cache = password.Password_Cache(self.config)
            action = None
            running = cache.status()

            if not running:
                logger.debug('pw-cache is not running', self)
                action = 'start'

            if running and not cache.checkVersion():
                logger.debug('pw-cache is running but is an old version', self)
                action = 'restart'

            bit = tools.which('backintime')

            if not action is None and not bit is None and len(bit):
                cmd = [bit, 'pw-cache', action]
                logger.debug(f'Call command: {cmd}', self)

                proc = subprocess.Popen(cmd,
                                        stdout = subprocess.DEVNULL,
                                        stderr = subprocess.DEVNULL)

                if proc.returncode:
                    logger.error(
                        f'Failed to {action} pw-cache: {proc.returncode}',
                        self)

                    pass

    def mount(self, mode=None, check=True, **kwargs):
        """High-level `mount`. Check if the selected ``mode`` need to be mounted,
        select the low-level backend and mount it.

        Args:
            mode (str): Mode to use. One of 'local', 'ssh', 'local_encfs' or
                'ssh_encfs'.
            check (bool): If ``True`` run :py:func:`MountControl.preMountCheck`
                before mounting.
            **kwargs: Keyword arguments paste to low-level
                :py:class:`MountControl` subclass backend.

        Returns:
            str: Hash ID used as mountpoint.

        Raises:
            exceptions.MountException: If a check failed.
            exceptions.HashCollision: If hash ID was used before but umount
                info wasn't identical.
        """
        self.config.PLUGIN_MANAGER.load(cfg=self.config)
        self.config.PLUGIN_MANAGER.mount(self.profile_id)

        if mode is None:
            mode = self.config.snapshotsMode(self.profile_id)

        if self.config.SNAPSHOT_MODES[mode][0] is None:
            # mode doesn't need to mount
            return 'local'

        else:

            while True:  # ???

                try:
                    mounttools = self.config.SNAPSHOT_MODES[mode][0]
                    backend = mounttools(cfg = self.config,
                                         profile_id = self.profile_id,
                                         tmp_mount = self.tmp_mount,
                                         mode = mode,
                                         parent = self.parent,
                                         **kwargs)

                    return backend.mount(check = check)

                except HashCollision as ex:
                    logger.warning(str(ex), self)
                    del backend
                    check = False

                    continue

                break

    def umount(self, hash_id=None):
        """High-level `unmount`. Unmount the low-level backend. This will read
        unmount infos written next to the mountpoint identified by ``hash_id``
        and unmount it.

        Args:
            hash_id (bool): Hash ID used as mountpoint before that should get
                            unmounted.

        Raises:
            exceptions.MountException: If a check failed.
        """
        self.config.PLUGIN_MANAGER.load(cfg=self.config)
        self.config.PLUGIN_MANAGER.unmount(self.profile_id)

        if hash_id is None:
            hash_id = self.config.current_hash_id

        if hash_id == 'local':
            # mode doesn't need to umount
            return

        umount_info = os.path.join(
            self.config._LOCAL_MOUNT_ROOT, hash_id, 'umount')

        with open(umount_info, 'r') as f:
            data_string = f.read()
            f.close()

        kwargs = json.loads(data_string)
        mode = kwargs.pop('mode')
        mounttools = self.config.SNAPSHOT_MODES[mode][0]
        backend = mounttools(cfg=self.config,
                             profile_id=self.profile_id,
                             tmp_mount=self.tmp_mount,
                             mode=mode,
                             hash_id=hash_id,
                             parent=self.parent,
                             **kwargs)

        backend.umount()

    def preMountCheck(self, mode=None, first_run=False, **kwargs):
        """
        High-level check. Run :py:func:`MountControl.preMountCheck` to check
        if all conditions for :py:func:`Mount.mount` are set.

        Should be called with ``first_run = True`` to check if new settings are
        correct before saving them.

        Args:
            mode (str): Mode to use. One of 'local', 'ssh', 'local_encfs' or
                'ssh_encfs'.
            first_run (bool): Run intense checks that only need to run after
                changing settings but not every time before mounting.
            **kwargs: Keyword arguments paste to low-level
                :py:class:`MountControl` subclass backend.

        Returns:
            bool: ``True`` if all checks where okay.

        Raises:
            exceptions.MountException: If a check failed.
        """
        if mode is None:
            mode = self.config.snapshotsMode(self.profile_id)

        # sshtools.SSH, encfstools.EncFS_mount, encfstools.EncFS_SSH
        Mounttools = self.config.SNAPSHOT_MODES[mode][0]

        # "local" mode
        if Mounttools is None:
            # mode doesn't need to mount
            return True

        backend = Mounttools(cfg=self.config,
                             profile_id=self.profile_id,
                             tmp_mount=self.tmp_mount,
                             mode=mode,
                             parent=self.parent,
                             **kwargs)

        return backend.preMountCheck(first_run)

    def remount(self, new_profile_id, mode = None, hash_id = None, **kwargs):
        """
        High-level `remount`. Unmount the old profile presented by ``hash_id``
        and mount new profile ``new_profile_id`` with mode ``mode``. If old and
        new mounts are the same just add new symlinks and keep the mount.

        Args map to profiles::

            new_profile_id          <= new profile
            mode                    <= new profile
            kwargs                  <= new profile
            hash_id                 <= old profile
            self.profile_id         <= old profile

        Args:
            new_profile_id (str):   Profile ID that should get mounted
            mode (str):             mode to use for new mount. One of 'local',
                                    'ssh', 'local_encfs' or 'ssh_encfs'
            hash_id (str):          Hash ID used as mountpoint on the old mount,
                                    that should get unmounted
            **kwargs:               keyword arguments paste to low-level
                                    :py:class:`MountControl` subclass backend
                                    for the new mount

        Returns:
            str:                    Hash ID used as mountpoint

        Raises:
            exceptions.MountException:
                                    if a check failed
            exceptions.HashCollision:
                                    if Hash ID was used before but umount info
                                    wasn't identical
        """
        if mode is None:
            mode = self.config.snapshotsMode(new_profile_id)
        if hash_id is None:
            hash_id = self.config.current_hash_id

        if self.config.SNAPSHOT_MODES[mode][0] is None:
            #new profile don't need to mount.
            self.umount(hash_id = hash_id)
            return 'local'

        if hash_id == 'local':
            #old profile don't need to umount.
            self.profile_id = new_profile_id
            return self.mount(mode = mode, **kwargs)

        mounttools = self.config.SNAPSHOT_MODES[mode][0]
        backend = mounttools(cfg = self.config,
                             profile_id = new_profile_id,
                             tmp_mount = self.tmp_mount,
                             mode = mode,
                             parent = self.parent,
                             **kwargs)
        if backend.compareRemount(hash_id):
            #profiles uses the same settings. just swap the symlinks
            backend.removeSymlink(profile_id = self.profile_id)
            backend.setSymlink(profile_id = new_profile_id, hash_id = hash_id)
            return hash_id
        else:
            #profiles are different. we need to umount and mount again
            self.umount(hash_id = hash_id)
            self.profile_id = new_profile_id
            return self.mount(mode = mode, **kwargs)

class MountControl:
    """This is the low-level mount API. This should be subclassed by backends.

    Subclasses should have its own ``__init__`` but **must** also call the
    inherited ``__init__``. See module description (:py:mod:`mount`) for
    a detailed example.

    You **must** overwrite methods:

    - :py:func:`MountControl._mount`

    You **can** overwrite methods:

    - :py:func:`MountControl._umount`
    - :py:func:`MountControl.preMountCheck`
    - :py:func:`MountControl.postMountCheck`
    - :py:func:`MountControl.preUmountCheck`
    - :py:func:`MountControl.postUmountCheck`

    These arguments (all of type :py:obj:`str`) **must** be defined in
    ``self`` namespace by subclassing ``__init__`` method:

    - ``mountproc``: process used to mount
    - ``log_command``: shortened form of mount command used in logs
    - ``symlink_subfolder``: mountpoint-subfolder which should be linked

    Args:
        cfg (config.Config):    current config
        profile_id (str):       profile ID that should be used
        hash_id (str):          crc32 hash used to identify identical mountpoints
        tmp_mount (bool):       if ``True`` mount to a temporary destination
        parent (QWidget):       parent widget for QDialogs or ``None`` if there
                                is no parent
        symlink (bool):         if ``True`` set symlink to mountpoint
        mode (str):             one of ``local``, ``local_encfs``, ``ssh`` or
                                ``ssh_encfs``
        hash_collision (int):   global value used to prevent hash collisions on
                                mountpoints
    """

    def __init__(self,
                 cfg=None,
                 profile_id=None,
                 hash_id=None,
                 tmp_mount=False,
                 parent=None,
                 symlink=True,
                 *args,
                 **kwargs):
        # The following members should get valid values from the inheriting
        # class.
        self.mode = None
        self.log_command = None
        self.mountproc = None
        self.symlink_subfolder = None

        self.config = cfg or config.Config()

        self.profile_id = profile_id or self.config.currentProfile()

        self.tmp_mount = tmp_mount
        self.hash_id = hash_id
        self.parent = parent
        self.symlink = symlink

        self.local_host = self.config.host()
        self.local_user = getpass.getuser()
        self.pid = str(os.getpid())

        self.all_kwargs = {}

        self.setattrKwargs(
            'mode', self.config.snapshotsMode(self.profile_id), **kwargs)
        self.setattrKwargs(
            'hash_collision', self.config.hashCollision(), **kwargs)

    def setDefaultArgs(self):
        """
        Set some arguments which are necessary for all backends.
        ``self.all_kwargs`` need to be filled through :py:func:`setattrKwargs`
        before calling this.
        """
        # self.destination should contain all arguments that are necessary for
        # mount.
        args = list(self.all_kwargs.keys())
        self.destination = '%s:' % self.all_kwargs['mode']

        args.remove('mode')
        args.sort()

        for arg in args:
            self.destination += ' %s' % self.all_kwargs[arg]

        # Unique id for every different mount settings. Similar settings even
        # in different profiles will generate the same hash_id and so share
        # the same mountpoint.
        if self.hash_id is None:
            self.hash_id = self.hash(self.destination)

        # e.g. ~/.local/share/backintime/mnt
        self.mount_root = self.config._LOCAL_MOUNT_ROOT
        self.snapshots_path = self.config.snapshotsPath(
            profile_id=self.profile_id,
            mode=self.mode,
            tmp_mount=self.tmp_mount)

        self.hash_id_path = self.hashIdPath()
        self.currentMountpoint = self.mountpoint()
        self.lock_path = self.lockPath()
        self.umount_info = self.umountInfoPath()

    def mount(self, check=True):
        """
        Low-level `mount`. Set mountprocess lock and prepare mount, run checks
        and than call :py:func:`_mount` for the subclassed backend. Finally set
        mount lock and symlink and release mountprocess lock.

        Args:
            check (bool): If ``True`` run :py:func:`preMountCheck` before
                mounting.

        Returns:
            str: Hash ID used as mountpoint.

        Raises:
            exceptions.MountException: If a check failed.
            exceptions.HashCollision: If Hash ID was used before but umount
                info wasn't identical.
        """
        self.createMountStructure()
        self.mountProcessLockAcquire()

        try:
            if self.mounted():

                if not self.compareUmountInfo():
                    # We probably have a hash collision
                    self.config.incrementHashCollision()
                    raise HashCollision(
                        f'Hash collision occurred in hash_id {self.hash_id}. '
                        'Incrementing global value hash_collision and '
                        'trying again.')

                logger.info('Mountpoint {} is already mounted'
                            .format(self.currentMountpoint),
                            self)
            else:
                if check:
                    self.preMountCheck()

                self._mount()
                self.postMountCheck()

                logger.info(
                    f'mount {self.log_command} on {self.currentMountpoint}',
                    self)
                self.writeUmountInfo()

        except Exception:
            # ???
            raise

        else:
            self.mountLockAquire()
            self.setSymlink()

        finally:
            self.mountProcessLockRelease()

        return self.hash_id

    def umount(self):
        """
        Low-level `umount`. Set mountprocess lock, run umount checks and call
        :py:func:`_umount` for the subclassed backend. Finally release
        mount lock, remove symlink and release mountprocess lock.

        Raises:
            exceptions.MountException:  if a check failed
        """
        self.mountProcessLockAcquire()

        msg_begin = f'Mountpoint {self.currentMountpoint} '

        try:
            if not os.path.isdir(self.hash_id_path):
                logger.info(msg_begin + 'does not exist.', self)

            else:
                if not self.mounted():
                    logger.info(msg_begin + 'is not mounted.', self)

                else:
                    if self.mountLockCheck():
                        logger.info(msg_begin + 'still in use. Keep mounted.',
                                    self)

                    else:
                        self.preUmountCheck()
                        self._umount()
                        self.postUmountCheck()

                        if os.listdir(self.currentMountpoint):
                            logger.warning(
                                msg_begin + 'not empty after unmount', self)

                        else:
                            logger.info(
                                f'unmount {self.log_command} '
                                f'from {self.currentMountpoint}',
                                self)

        except Exception:
            raise

        else:
            self.mountLockRelease()
            self.removeSymlink()

        finally:
            self.mountProcessLockRelease()

    def _mount(self):
        """
        Backend mount method. This **must** be overwritten in the backend which
        subclasses :py:class:`MountControl`.
        """
        raise NotImplementedError('_mount need to be overwritten in backend')

    def _umount(self):
        """
        Unmount with ``fusermount -u`` for fuse based backends. This **can** be
        overwritten by backends which subclasses :py:class:`MountControl`.

        Raises:
            exceptions.MountException: If unmount failed.
        """
        try:
            subprocess.check_call(['fusermount', '-u', self.currentMountpoint])

        except subprocess.CalledProcessError as exc:
            raise MountException(
                _('Unable to unmount {mountprocess} from {mountpoint}.')
                .format(mountprocess=self.mountproc,
                        mountpoint=self.currentMountpoint)) from exc

    def preMountCheck(self, first_run=False):
        """
        Check what ever conditions must be given for the mount to be done
        successful. This **can** be overwritten in backends which
        subclasses :py:class:`MountControl`.

        Returns:
            bool:       ``True`` if all checks where okay

        Raises:
            exceptions.MountException:
                        if backend can not mount

        Note:
            This can also be used to prepare things before running
            :py:func:`_mount`
        """
        return True

    def postMountCheck(self):
        """
        Check if the mount was successful. This **can** be overwritten in
        backends which subclasses :py:class:`MountControl`.

        Returns:
            bool:       ``True`` if all checks where okay

        Raises:
            exceptions.MountException:
                        if backend wasn't mount successful

        Note:
            This can also be used to clean up after running :py:func:`_mount`
        """
        return True

    def preUmountCheck(self):
        """
        Check if backend is safe to umount. This **can** be overwritten in
        backends which subclasses :py:class:`MountControl`.

        Returns:
            bool:       ``True`` if all checks where okay

        Raises:
            exceptions.MountException:
                        if backend can not umount

        Note:
            This can also be used to prepare things before running
            :py:func:`_umount`
        """
        return True

    def postUmountCheck(self):
        """
        Check if unmount was successful. This **can** be overwritten in backends
        which subclasses :py:class:`MountControl`.

        Returns:
            bool:       ``True`` if all checks where okay

        Raises:
            exceptions.MountException:
                        if backend wasn't unmounted successful

        Note:
            This can also be used to clean up after running :py:func:`_umount`
        """
        return True

    def checkFuse(self):
        """
        Check if command in ``self.mountproc`` is installed.

        Raises:
            exceptions.MountException: If either command is not available.
        """

        if not tools.checkCommand(self.mountproc):
            logger.debug(f'{self.mountproc} is missing', self)

            raise MountException(
                _('{command} not found. Please install it '
                  '(e.g. via "{installcommand}")').format(
                      command=self.mountproc,
                      installcommand=f"'apt-get install {self.mountproc}'")
            )

    def mounted(self):
        """
        Check if the mountpoint is already mounted.

        Returns:
            bool:   ``True`` if mountpoint is mounted

        Raises:
            exceptions.MountException:
                    if mountpoint is not mounted but also not empty
        """
        if os.path.ismount(self.currentMountpoint):
            return True

        else:
            try:
                if os.listdir(self.currentMountpoint):
                    raise MountException(
                        _('Mountpoint {mntpoint} not empty.').format(
                            mntpoint=self.currentMountpoint))

            except FileNotFoundError:
                pass

            return False

    def createMountStructure(self):
        """
        Create folders that are necessary for mounting.

        Folder structure in ``~/.local/share/backintime/mnt/``
        (``self.mount_root``)::

            .
            ├── <pid>.lock              <=  mountprocess lock that will prevent
            │                               different processes modifying
            │                               mountpoints at one time
            │
            ├── <hash_id>/              <=  ``self.hash_id_path`` will be
            │   │                           shared by all profiles with the
            │   │                           same mount settings
            │   │
            │   ├── mountpoint/         <=  ``self.currentMountpoint`` real
            │   │                           mountpoint
            │   │
            │   ├── umount              <=  ``self.umount_info`` json file with
            │   │                           all necessary args for unmount
            │   │
            │   └── locks/              <=  ``self.lock_path`` for each process
            │                               you have a ``<pid>.lock`` file
            │
            ├── <profile id>_<pid>/     <=  sym-link to the right path. return
            │                               by config.snapshotsPath (can be
            │                               ../mnt/<hash_id>/mount_point for ssh
            │                               or ../mnt/<hash_id>/<HOST>/<SHARE>
            │                               for fusesmb ...)
            │
            └── tmp_<profile id>_<pid>/ <=  sym-link for testing mountpoints
                                            in settingsdialog
        """
        tools.mkdir(self.mount_root, 0o700)
        tools.mkdir(self.hash_id_path, 0o700)
        tools.mkdir(self.currentMountpoint, 0o700, False)
        tools.mkdir(self.lock_path, 0o700)

    def mountProcessLockAcquire(self, timeout=60):
        """
        Create a short term lock only for blocking other processes changing
        mounts at the same time.

        Args:
            timeout (int): Wait ``timeout`` seconds before fail acquiring
                the lock.

        Raises:
            exceptions.MountException: If timed out.
        """
        lock_path = self.mount_root
        lockSuffix = '.lock'
        # e.g. ~/.local/share/backintime/mnt/123456.lock
        lock = os.path.join(lock_path, self.pid + lockSuffix)

        # Every second
        count = 0
        while self.checkLocks(lock_path, lockSuffix):
            count += 1

            if count == timeout:
                raise MountException('Mountprocess lock timeout')

            sleep(1)

        logger.debug(f'Acquire mountprocess lock {lock}', self)

        with open(lock, 'w') as f:
            f.write(self.pid)

    def mountProcessLockRelease(self):
        """Remove mountprocess lock."""
        lock_path = self.mount_root
        lockSuffix = '.lock'
        lock = os.path.join(lock_path, self.pid + lockSuffix)

        logger.debug(f'Release mountprocess lock {lock}', self)

        if os.path.exists(lock):
            os.remove(lock)

    def mountLockAquire(self):
        """
        Create a lock file for a mountpoint to prevent unmounting as long as
        this process is running.
        """
        lockSuffix = '.tmp.lock' if self.tmp_mount else '.lock'
        lock = os.path.join(self.lock_path, self.pid + lockSuffix)

        logger.debug(f'Set mount lock {lock}', self)

        with open(lock, 'w') as f:
            f.write(self.pid)

    def mountLockCheck(self):
        """
        Check for locks on the current mountpoint.

        Returns:
            bool:   ``True`` if there are any locks
        """
        lockSuffix = '.lock'
        return self.checkLocks(self.lock_path, lockSuffix)

    def mountLockRelease(self):
        """
        Remove mountpoint lock for this process.
        """
        lockSuffix = '.tmp.lock' if self.tmp_mount else '.lock'
        lock = os.path.join(self.lock_path, self.pid + lockSuffix)

        if os.path.exists(lock):
            logger.debug(f'Remove mount lock {lock}', self)
            os.remove(lock)

    def checkLocks(self, path, lock_suffix):
        """Check existence of active and foreign locks.

        The lock owning process is specified by the PID contained in the
        filename of the lock file used. Lock files of the current process are
        ignored and ``False`` is returned if they share the same tmp-mount
        state. If a lock exist but its process not the lock is removed and
        ``False`` returned. In that latter case mount symlinks related to that
        lock also removed.

        Args:
            path (str): Full path to lock directory.
            lock_suffix (str): Last part of locks name.

        Returns:
            bool: ``True`` if there are active locks in ``path``.

        Raises:
            FileNotFoundError: If the path does not exists.

        """
        if isinstance(path, str):
            path = Path(path)

        for lock_fp in path.iterdir():

            # Not a lock file?
            if lock_fp.suffix != lock_suffix:
                # next file
                continue

            # Secondary suffix is "tmp"? (e.g. "12345.tmp.lock")
            is_tmp = lock_fp.with_suffix('').suffix == '.tmp'

            # Extract PID from lock file name
            lock_pid = lock_fp.stem
            if is_tmp:
                lock_pid = lock_pid[:-4]  # cut ".tmp" from the end

            # Ignore process's own lock files.
            if lock_pid == self.pid:
                # ...with the same tmp-state.
                if is_tmp == self.tmp_mount:
                    # Dev note (buhtz, 2024-11-22): I have no idea what makes
                    # a mount temporary.
                    continue

            if tools.processAlive(int(lock_pid)):
                return True

            logger.debug(f'Remove old and invalid lock {lock_fp}', self)

            # Clean up the lock file
            lock_fp.unlink()

            # Clean up related symlinks
            for symlink in Path(self.mount_root).iterdir():
                if symlink.name.endswith(f'_{lock_pid}'):
                    symlink.unlink()

        return False

    def setattrKwargs(self, arg, default, store=True, **kwargs):
        """
        Set attribute ``arg`` in local namespace (self.arg). Also collect all
        args in ``self.all_kwargs`` which will be hashed later and used as
        mountpoint name and also be written as unmount_info.

        Args:
            arg (str):      argument name
            default:        default value used if ``arg`` is not in ``kwargs``
            store (bool):   if ``True`` add ``arg`` to ``self.all_kwargs``
            **kwargs:       arguments given on backend constructor
        """
        if arg in kwargs:
            value = kwargs[arg]
        else:
            value = default
        setattr(self, arg, value)
        if store:
            #make dictionary with all used args for umount
            self.all_kwargs[arg] = value

    def writeUmountInfo(self):
        """
        Write content of ``self.all_kwargs`` to file
        ``~/.local/share/backintime/mnt/<hash_id>/umount``.
        This will be used to unmount the filesystem later.
        """
        data_string = json.dumps(self.all_kwargs)
        with open(self.umount_info, 'w') as f:
            f.write(data_string)
            f.close

    def readUmountInfo(self, umount_info = None):
        """
        Read keyword arguments from file ``umount_info``.

        Args:
            umount_info (str):  full path to <hash_id>/umount file. If ``None``
                                current ``<hash_id>/umount`` file will be used

        Returns:
            dict:               previously written ``self.all_kwargs``
        """
        if umount_info is None:
            umount_info = self.umount_info
        with open(umount_info, 'r') as f:
            data_string = f.read()
            f.close()
        return json.loads(data_string)

    def compareUmountInfo(self, umount_info = None):
        """
        Compare current ``self.all_kwargs`` with those from file ``umount_info``.

        This should prevent hash collisions of two different mounts.

        Args:
            umount_info (str):  full path to <hash_id>/umount file

        Returns:
            bool:               ``True`` if ``self.all_kwargs`` and ``kwargs``
                                read from ``umount_info`` file are identiacal
        """
        #run self.all_kwargs through json first
        current_kwargs = json.loads(json.dumps(self.all_kwargs))
        saved_kwargs = self.readUmountInfo(umount_info)
        if not len(current_kwargs) == len(saved_kwargs):
            return False
        for arg in list(current_kwargs.keys()):
            if not arg in list(saved_kwargs.keys()):
                return False
            if not current_kwargs[arg] == saved_kwargs[arg]:
                return False
        return True

    def compareRemount(self, old_hash_id):
        """
        Compare mount arguments between current and ``old_hash_id``. If they are
        identical we could reuse the mount and don't need to remount.

        Args:
            old_hash_id (str):  Hash ID of the old mountpoint

        Returns:
            bool:               True if the old mountpoint and current are
                                identiacal
        """
        if old_hash_id == self.hash_id:
            return self.compareUmountInfo(self.umountInfoPath(old_hash_id))
        return False

    def setSymlink(self, profile_id = None, hash_id = None, tmp_mount = None):
        """
        If ``self.symlink`` is ``True`` set symlink
        ``~/.local/share/backintime/mnt/<profile id>_<pid>``. Target will be
        either the mountpoint or a subfolder of the mountpoint if
        ``self.symlink_subfolder`` is set.

        Args:
            profile_id (str):   Profile ID that should be linked. If ``None``
                                use ``self.profile_id``
            hash_id (str):      Hash ID of mountpoint where this sysmlink should
                                point to. If ``None`` use ``self.hash_id``
            tmp_mount (bool):   Set a temporary symlink just for testing new
                                settings
        """
        if not self.symlink:
            return

        if profile_id is None:
            profile_id = self.profile_id

        if hash_id is None:
            hash_id = self.hash_id

        if tmp_mount is None:
            tmp_mount = self.tmp_mount

        dst = self.config.snapshotsPath(profile_id=profile_id,
                                        mode=self.mode,
                                        tmp_mount=tmp_mount)
        mountpoint = self.mountpoint(hash_id)

        if self.symlink_subfolder is None:
            src = mountpoint
        else:
            src = os.path.join(mountpoint, self.symlink_subfolder)

        if os.path.exists(dst):
            os.remove(dst)

        os.symlink(src, dst)

    def removeSymlink(self, profile_id=None, tmp_mount=None):
        """
        Remove symlink ``~/.local/share/backintime/mnt/<profile id>_<pid>``.

        Args:
            profile_id (str): Profile ID for the symlink.
            tmp_mount (bool): Symlink is a temporary link for testing new
                settings.
        """
        if not self.symlink:
            return

        if profile_id is None:
            profile_id = self.profile_id

        if tmp_mount is None:
            tmp_mount = self.tmp_mount

        os.remove(self.config.snapshotsPath(
            profile_id=profile_id,
            mode=self.mode,
            tmp_mount=tmp_mount))

    def hash(self, s):
        """
        Create a CRC32 hash of string ``s``.

        Args:
            s (str):    string that should be hashed

        Returns:
            str:        hash of string ``s``
        """
        return '%X' % (crc32(s.encode()) & 0xFFFFFFFF)

    def hashIdPath(self, hash_id = None):
        """
        Get path ``~/.local/share/backintime/mnt/<hash_id>``.

        Args:
            hash_id (str):  Unique identifier for a mountpoint. If ``None`` use
                            ``self.hash_id``

        Returns:
            str:            full path to ``<hash_id>``
        """
        if hash_id is None:
            hash_id = self.hash_id
        return os.path.join(self.mount_root, self.hash_id)

    def mountpoint(self, hash_id = None):
        """
        Get path ``~/.local/share/backintime/mnt/<hash_id>/mountpoint``.

        Args:
            hash_id (str):  Unique identifier for a mountpoint

        Returns:
            str:            full path to ``<hash_id>/mountpoint``
        """
        return os.path.join(self.hashIdPath(hash_id), 'mountpoint')

    def lockPath(self, hash_id = None):
        """
        Get path ``~/.local/share/backintime/mnt/<hash_id>/locks``.

        Args:
            hash_id (str):  Unique identifier for a mountpoint

        Returns:
            str:            full path to ``<hash_id>/locks```
        """
        return os.path.join(self.hashIdPath(hash_id), 'locks')

    def umountInfoPath(self, hash_id = None):
        """
        Get path ``~/.local/share/backintime/mnt/<hash_id>/umount``.

        Args:
            hash_id (str):  Unique identifier for a mountpoint

        Returns:
            str:            full path to ``<hash_id>/umount```
        """
        return os.path.join(self.hashIdPath(hash_id), 'umount')



# SPDX-FileCopyrightText: © 2008-2022 Oprea Dan
# SPDX-FileCopyrightText: © 2008-2022 Bart de Koning
# SPDX-FileCopyrightText: © 2008-2022 Richard Bailey
# SPDX-FileCopyrightText: © 2008-2022 Germar Reitze
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import (QDialog,
                             QLabel,
                             QPlainTextEdit,
                             QVBoxLayout,
                             QHBoxLayout,
                             QComboBox,
                             QDialogButtonBox,
                             QCheckBox,
                             )
from PyQt6.QtCore import QFileSystemWatcher
import qttools
import snapshots
import encfstools
import snapshotlog
import tools
import qttools
from statedata import StateData


class LogViewDialog(QDialog):
    def __init__(self, parent, sid=None, systray=False):
        """
        Instantiate a snapshot log file viewer

        Args:
            parent:
            sid (:py:class:`SID`): snapshot ID whose log file shall be shown
                                   (``None`` = show last log)
            systray (bool): TODO Show log from systray icon or from App (boolean)
        """
        if systray:
            super(LogViewDialog, self).__init__()
        else:
            super(LogViewDialog, self).__init__(parent)

        self.config = parent.config
        self.snapshots = parent.snapshots
        self.mainWindow = parent
        self.sid = sid
        self.enableUpdate = False
        self.decode = None

        state_data = StateData()
        self.resize(*state_data.logview_dims)

        import icon
        self.setWindowIcon(icon.VIEW_SNAPSHOT_LOG)
        if self.sid is None:
            self.setWindowTitle(_('Last Log View'))
        else:
            self.setWindowTitle(_('Snapshot Log View'))

        self.mainLayout = QVBoxLayout(self)

        layout = QHBoxLayout()
        self.mainLayout.addLayout(layout)

        # profiles
        self.lblProfile = QLabel(_('Profile:'), self)
        layout.addWidget(self.lblProfile)

        self.comboProfiles = qttools.ProfileCombo(self)
        layout.addWidget(self.comboProfiles, 1)
        self.comboProfiles.currentIndexChanged.connect(self.profileChanged)

        # snapshots
        self.lblSnapshots = QLabel(_('Snapshots:'), self)
        layout.addWidget(self.lblSnapshots)
        self.comboSnapshots = qttools.SnapshotCombo(self)
        layout.addWidget(self.comboSnapshots, 1)
        self.comboSnapshots.currentIndexChanged.connect(self.comboSnapshotsChanged)

        if self.sid is None:
            self.lblSnapshots.hide()
            self.comboSnapshots.hide()

        if self.sid or systray:
            self.lblProfile.hide()
            self.comboProfiles.hide()

        # filter
        layout.addWidget(QLabel(_('Filter:'), self))

        self.comboFilter = QComboBox(self)
        layout.addWidget(self.comboFilter, 1)
        self.comboFilter.currentIndexChanged.connect(self.comboFilterChanged)

        self.comboFilter.addItem(_('All'), snapshotlog.LogFilter.NO_FILTER)

        # Note about ngettext plural forms: n=102 means "Other" in Arabic and
        # "Few" in Polish.
        # Research in translation community indicate this as the best fit to
        # the meaning of "all".
        self.comboFilter.addItem(
            ' + '.join((_('Errors'), _('Changes'))),
            snapshotlog.LogFilter.ERROR_AND_CHANGES)
        self.comboFilter.setCurrentIndex(self.comboFilter.count() - 1)
        self.comboFilter.addItem(_('Errors'), snapshotlog.LogFilter.ERROR)
        self.comboFilter.addItem(_('Changes'), snapshotlog.LogFilter.CHANGES)
        self.comboFilter.addItem(ngettext('Information', 'Information', 2),
                                 snapshotlog.LogFilter.INFORMATION)
        self.comboFilter.addItem(
            _('rsync transfer failures (experimental)'),
            snapshotlog.LogFilter.RSYNC_TRANSFER_FAILURES)

        # text view
        self.txtLogView = QPlainTextEdit(self)
        self.txtLogView.setFont(QFont('Monospace'))
        self.txtLogView.setReadOnly(True)
        self.txtLogView.setLineWrapMode(QPlainTextEdit.LineWrapMode.NoWrap)
        self.mainLayout.addWidget(self.txtLogView)

        #
        self.mainLayout.addWidget(
            QLabel(_('[E] Error, [I] Information, [C] Change')))

        # decode path
        self.cbDecode = QCheckBox(_('decode paths'), self)
        self.cbDecode.stateChanged.connect(self.cbDecodeChanged)
        self.mainLayout.addWidget(self.cbDecode)

        # buttons
        buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Close)
        self.mainLayout.addWidget(buttonBox)
        buttonBox.rejected.connect(self.close)

        self.updateSnapshots()
        self.updateDecode()
        self.updateProfiles()

        # watch for changes in log file
        self.watcher = QFileSystemWatcher(self)
        if self.sid is None:
            # only watch if we show the last log
            log = self.config.takeSnapshotLogFile(
                self.comboProfiles.currentProfileID())
            self.watcher.addPath(log)
        # passes the path to the changed file to updateLog()
        self.watcher.fileChanged.connect(self.updateLog)

    def cbDecodeChanged(self):
        if self.cbDecode.isChecked():
            if not self.decode:
                self.decode = encfstools.Decode(self.config)

        else:
            if self.decode is not None:
                self.decode.close()
            self.decode = None

        self.updateLog()

    def profileChanged(self, index):
        if not self.enableUpdate:
            return
        profile_id = self.comboProfiles.currentProfileID()
        self.mainWindow.comboProfiles.setCurrentProfileID(profile_id)
        self.mainWindow.comboProfileChanged(None)

        self.updateDecode()
        self.updateLog()

    def comboSnapshotsChanged(self, index):
        if not self.enableUpdate:
            return
        self.sid = self.comboSnapshots.currentSnapshotID()
        self.updateLog()

    def comboFilterChanged(self, index):
        self.updateLog()

    def updateProfiles(self):
        current_profile_id = self.config.currentProfile()

        self.comboProfiles.clear()

        qttools.update_combo_profiles(self.config, self.comboProfiles, current_profile_id)

        self.enableUpdate = True
        self.updateLog()

        if len(self.config.profilesSortedByName()) <= 1:
            self.lblProfile.setVisible(False)
            self.comboProfiles.setVisible(False)

    def updateSnapshots(self):
        if self.sid:
            self.comboSnapshots.clear()
            for sid in snapshots.iterSnapshots(self.config):
                self.comboSnapshots.addSnapshotID(sid)
                if sid == self.sid:
                    self.comboSnapshots.setCurrentSnapshotID(sid)

    def updateDecode(self):
        if self.config.snapshotsMode() == 'ssh_encfs':
            self.cbDecode.show()
        else:
            self.cbDecode.hide()
            if self.cbDecode.isChecked():
                self.cbDecode.setChecked(False)

    def updateLog(self, watchPath=None):
        """
        Show the log file of the current snapshot in the GUI

        Args:
            watchPath: FQN to a log file (as string) whose changes are watched
                       via ``QFileSystemWatcher``. In case of changes
                       this function is called with the log file
                       and only the new lines in the log file are appended
                       to the log file widget in the GUI
                       Use ``None`` if a complete log file shall be shown
                       at once.
        """
        if not self.enableUpdate:
            return

        mode = self.comboFilter.itemData(self.comboFilter.currentIndex())

        # TODO This expressions is hard to understand (watchPath is not a
        # boolean!)
        if watchPath and self.sid is None:
            # remove path from watch to prevent multiple updates at the same
            # time
            self.watcher.removePath(watchPath)
            # append only new lines to txtLogView
            log = snapshotlog.SnapshotLog(
                self.config, self.comboProfiles.currentProfileID())
            for line in log.get(mode=mode,
                                decode=self.decode,
                                skipLines=self.txtLogView.document().lineCount()-1):
                self.txtLogView.appendPlainText(line)

            # re-add path to watch after 5sec delay
            alarm = tools.Alarm(
                callback=lambda: self.watcher.addPath(watchPath),
                overwrite=False)
            alarm.start(5)

        elif self.sid is None:
            log = snapshotlog.SnapshotLog(
                self.config, self.comboProfiles.currentProfileID())
            self.txtLogView.setPlainText(
                '\n'.join(log.get(mode=mode, decode=self.decode)))

        else:
            self.txtLogView.setPlainText(
                '\n'.join(self.sid.log(mode, decode=self.decode)))

    def closeEvent(self, event):
        state_data = StateData()
        state_data.logview_dims = (self.width(), self.height())
        event.accept()



# SPDX-FileCopyrightText: © 2025 Christian Buhtz <c.buhtz@posteo.jp>
#
# SPDX-License-Identifier: GPL-2.0-or-later
#
# This file is part of the program "Back In Time" which is released under GNU
# General Public License v2 (GPLv2). See LICENSES directory or go to
# <https://spdx.org/licenses/GPL-2.0-or-later.html>.
"""Tests about statefile module."""
# pylint: disable=wrong-import-position,wrong-import-order
import unittest
from qttools_path import registerBackintimePath
registerBackintimePath('common')
import statedata  # noqa: E402


class IsSingleton(unittest.TestCase):
    """StateData instance is a singleton."""

    @classmethod
    def tearDownClass(cls):
        # Delete existing StateData instance
        try:
            # pylint: disable-next=protected-access
            del statedata.StateData._instances[statedata.StateData]
        except KeyError:
            pass

    def setUp(self):
        # Clean up all instances
        try:
            # pylint: disable-next=protected-access
            del statedata.StateData._instances[statedata.StateData]
        except KeyError:
            pass

    def test_identity(self):
        """Identical identity."""
        one = statedata.StateData()
        two = statedata.StateData()

        self.assertEqual(id(one), id(two))

    def test_content(self):
        """Identical values."""
        one = statedata.StateData()
        two = statedata.StateData()

        one['foobar'] = 7

        self.assertEqual(one, two)


class Properties(unittest.TestCase):
    """Property access without errors."""

    @classmethod
    def tearDownClass(cls):
        # Delete existing StateData instance
        try:
            # pylint: disable-next=protected-access
            del statedata.StateData._instances[statedata.StateData]
        except KeyError:
            pass

    def setUp(self):
        # Delete existing StateData instance
        try:
            # pylint: disable-next=protected-access
            del statedata.StateData._instances[statedata.StateData]
        except KeyError:
            pass

    def test_read_empty_global(self):
        """Read properties from empty state data"""
        sut = statedata.StateData()

        self.assertEqual(sut.msg_release_candidate, None)
        self.assertEqual(sut.msg_encfs_global, False)
        self.assertEqual(sut.mainwindow_show_hidden, False)
        self.assertEqual(sut.files_view_sorting, (0, 0))
        self.assertEqual(sut.mainwindow_main_splitter_widths, (150, 450))
        self.assertEqual(sut.mainwindow_second_splitter_widths, (150, 300))

        with self.assertRaises(KeyError):
            # pylint: disable=pointless-statement
            sut.mainwindow_coords
            sut.mainwindow_dims
            sut.logview_dims
            sut.files_view_col_widths

    def test_profile_not_exist(self):
        """Profile does not exists."""
        sut = statedata.StateData()
        profile = sut.profile(42)

        with self.assertRaises(KeyError):
            # pylint: disable=pointless-statement
            profile.last_path
